import os
import requests
from typing import Union

from . import client
from .utils import parse_json, get_all_results, PrevisionException, EventTuple
from .api_resource import ApiResource, UniqueResourceMixin
from .dataset import Dataset
from .prediction import DeploymentPrediction, ValidationPrediction


class Export(ApiResource, UniqueResourceMixin):

    """ An export

    Args:
        _id (str): Unique id of the export
        exporter_id (str): Unique exporter id on which to create the export
        dataset_id (str, optional): Unique dataset id on which to create the export
        prediction_id (str, optional): Unique prediction id on which to create the export
    """

    resource = 'exports'

    def __init__(self, _id, exporter_id: str, dataset_id: str = None, prediction_id: str = None,
                 status: str = None, **kwargs):
        """ Instantiate a new :class:`.Export` object to manipulate an export resource
        on the platform. """
        super().__init__(_id=_id)

        self.exporter_id = exporter_id
        self.dataset_id = dataset_id
        self.state = status
        self.other_params = kwargs

    @classmethod
    def list(cls, exporter_id: str):
        """ List all the available exports given an exporter id.

        Args:
            exporter_id: exporter id

        Returns:
            list(:class:`.Export`): The fetched export objects
        """
        end_point = '/exporters/{}/exports'.format(exporter_id)
        exports = get_all_results(client, end_point, method=requests.get)
        return [cls(**data) for data in exports]

    @classmethod
    def from_id(cls, _id: str):
        """Get an export from the instance by its unique id.

        Args:
            _id (str): Unique id of the export to retrieve

        Returns:
            :class:`.Export`: The fetched export

        Raises:
            PrevisionException: Any error while fetching data from the platform
                or parsing the result
        """
        resp_json = super()._from_id(_id=_id)
        return cls(**resp_json)

    @classmethod
    def _new(cls, exporter_id: str, prediction: Union[DeploymentPrediction, ValidationPrediction] = None,
             dataset: Dataset = None, file_path: str = None, encoding: str = None, separator: str = None,
             decimal: str = None, thousands: str = None, wait_for_export: bool = False, **kwargs):
        """ Create a new exporter object on the platform.

        Args:
            exporter_id (str): Unique exporter id on which to create the export
            prediction (:class:`.DeploymentPrediction`|:class:`.ValidationPrediction`, optional): prediction to upload
            dataset (:class:`.Dataset`, optional): Dataset to upload
            file_path (str, optional): Path of the file to upload
            encoding (str, optional): Encoding of the file to upload
            separator (str, optional): Separator of the file to upload
            decimal (str, optional): Decimal of the file to upload
            thousands (str, optional): Thousands of the file to upload
            wait_for_export (bool, optional): Wether to wait until the export is complete or not

        Returns:
            :class:`.Export`: The registered export object

        Raises:
            PrevisionException: Any error while uploading data to the platform
                or parsing the result
            Exception: For any other unknown error
        """
        data = {}

        if dataset is not None:
            request_url = '/exporters/{}/dataset/{}'.format(exporter_id, dataset._id)
            create_resp = client.request(request_url,
                                         method=requests.post,
                                         message_prefix='Export dataset')

        elif prediction is not None:
            request_url = '/exporters/{}/prediction/{}'.format(exporter_id, prediction._id)
            if isinstance(prediction, DeploymentPrediction):
                data['prediction_type'] = 'deployment'
            elif isinstance(prediction, ValidationPrediction):
                data['prediction_type'] = 'experiment'
            else:
                msg = 'prediction must be of type DeploymentPrediction or ValidationPrediction,'
                msg += f' got: {type(prediction)}'
                raise PrevisionException(msg)
            create_resp = client.request(request_url,
                                         data=data,
                                         method=requests.post,
                                         message_prefix='Export prediction')

        elif file_path is not None:
            if 'origin' in kwargs:
                data['origin'] = kwargs['origin']
            if 'pipeline_scheduled_run_id' in kwargs:
                data['pipeline_scheduled_run_id'] = kwargs['pipeline_scheduled_run_id']
            if encoding is not None:
                data['encoding'] = encoding
            if separator is not None:
                data['separator'] = separator
            if decimal is not None:
                data['decimal'] = decimal
            if thousands is not None:
                data['thousands'] = thousands
            files = {}
            with open(file_path, 'r') as f:
                for k, v in data.items():
                    files[k] = (None, v)
                files['file'] = (os.path.basename(file_path), f, 'text/csv')
                request_url = '/exporters/{}/file'.format(exporter_id)
                create_resp = client.request(request_url,
                                             data=data,
                                             files=files,
                                             method=requests.post,
                                             message_prefix='Export file')

        else:
            raise PrevisionException('Need to specify one of: dataset, prediction or file_path')

        create_resp = parse_json(create_resp)

        if wait_for_export:
            try:
                specific_url = '/{}/{}'.format('exports', create_resp['_id'])
                client.event_manager.wait_for_event(create_resp['_id'],
                                                    cls.resource,
                                                    EventTuple(
                                                        'EXPORT_UPDATE',
                                                        ('status', 'done'),
                                                        [('status', 'failed')]),
                                                    specific_url=specific_url)

            except PrevisionException:
                # Extract log if failure
                resp_json = super()._from_id(_id=create_resp['_id'])
                msg = f'Export failure.\nLogs: {resp_json["logs"]}\nurl: {request_url}'
                if data:
                    msg += f'\ndata: {data}'
                raise PrevisionException(msg)

            except Exception:
                raise

        return cls.from_id(create_resp['_id'])

    @classmethod
    def export_file(cls, exporter_id: str, file_path: str, encoding: str = None, separator: str = None,
                    decimal: str = None, thousands: str = None, wait_for_export: bool = False, **kwargs):
        """ Upload a CSV file using an exporter.

        Args:
            exporter_id (str): Unique exporter id on which to create the export
            file_path (str): Path of the file to upload
            encoding (str, optional): Encoding of the file to upload
            separator (str, optional): Separator of the file to upload
            decimal (str, optional): Decimal of the file to upload
            thousands (str, optional): Thousands of the file to upload
            wait_for_export (bool, optional): Wether to wait until the export is complete or not

        Returns:
            :class:`.Export`: The registered export object
        """
        return cls._new(exporter_id, file_path=file_path, encoding=encoding, separator=separator,
                        decimal=decimal, thousands=thousands, wait_for_export=wait_for_export, **kwargs)

    @classmethod
    def export_dataset(cls, exporter_id, dataset: Dataset, wait_for_export: bool = False):
        """ Upload a :class:`.Dataset` from the current active project using an exporter.

        Args:
            exporter_id (str): Unique exporter id on which to create the export
            dataset (:class:`.Dataset`): Dataset to upload
            wait_for_export (bool, optional): Wether to wait until the export is complete or not

        Returns:
            :class:`.Export`: The registered export object
        """
        return cls._new(exporter_id, dataset=dataset, wait_for_export=wait_for_export)

    @classmethod
    def export_prediction(cls, exporter_id, prediction: Union[DeploymentPrediction, ValidationPrediction],
                          wait_for_export: bool = False):
        """ Upload a :class:`.DeploymentPrediction` or a :class:`.ValidationPrediction`
        from the current active project using an exporter.

        Args:
            exporter_id (str): Unique exporter id on which to create the export
            prediction (:class:`.DeploymentPrediction`|:class:`.ValidationPrediction`): Prediction to upload
            wait_for_export (bool, optional): Wether to wait until the export is complete or not

        Returns:
            :class:`.Export`: The registered export object
        """
        return cls._new(exporter_id, prediction=prediction, wait_for_export=wait_for_export)



from enum import Enum
from typing import Dict, List, Union
from previsionio.dataset import Dataset
from .experiment_config import DataType, ExperimentConfig, TypeProblem, YesOrNo, YesOrNoOrAuto
from .utils import to_json
from .model import TextSimilarityModel
from .experiment_version import BaseExperimentVersion
import previsionio as pio


class ModelEmbedding(Enum):
    """Embedding models for Text Similarity"""
    TFIDF = 'tf_idf'
    """Term Frequency - Inverse Document Frequency"""
    Transformer = 'transformer'
    """Transformer"""
    TransformerFineTuned = 'transformer_fine_tuned'
    """fine tuned Transformer"""


class TextSimilarityModels(Enum):
    """Similarity search models for Text Similarity"""
    BruteForce = 'brute_force'
    """Brute force search"""
    ClusterPruning = 'cluster_pruning'
    """Cluster Pruning"""
    IVFOPQ = 'ivf_opq'
    """InVerted File system and Optimized Product Quantization"""
    HKM = 'hkm'
    """Hierarchical K-Means"""
    LSH = 'lsh'
    """Locality Sensitive Hashing"""


class TextSimilarityLang(Enum):
    Auto = 'auto'
    French = 'fr'
    English = 'en'


class Preprocessing(object):
    config = {}

    def __init__(self,
                 word_stemming: YesOrNo = YesOrNo.Yes,
                 ignore_stop_word: YesOrNoOrAuto = YesOrNoOrAuto.Auto,
                 ignore_punctuation: YesOrNo = YesOrNo.No):
        self.word_stemming = word_stemming
        self.ignore_stop_word = ignore_stop_word
        self.ignore_punctuation = ignore_punctuation


class ModelsParameters(ExperimentConfig):
    """ Training configuration that holds the relevant data for an experiment description:
    the wanted feature engineering, the selected models, the training speed...

    Args:
        preprocessing (Preprocessing, optional): Dictionary of the text preprocessings to be applied
            (only for "tf_idf" embedding model),

            - *word_stemming*: default to "yes"
            - *ignore_stop_word*: default to "auto", choice will be made depending on if the
              text descriptions contain full sentences or not
            - *ignore_punctuation*: default to "no".
        model_embedding (ModelEmbedding, optional): Name of the embedding model to be used
            (among: "tf_idf", "transformer", "transformer_fine_tuned").
        models (list(TextSimilarityModels), optional): Names of the searching models to be used (among:
            "brute_force", "cluster_pruning", "ivf_opq", "hkm", "lsh").
    """

    config = {
        'model_embedding': 'model_embedding',
        'preprocessing': 'preprocessing',
        'models': 'models'
    }

    def __init__(self,
                 model_embedding: ModelEmbedding = ModelEmbedding.TFIDF,
                 preprocessing: Preprocessing = Preprocessing(),
                 models: List[TextSimilarityModels] = [TextSimilarityModels.BruteForce]):
        self.model_embedding = model_embedding
        if isinstance(preprocessing, Preprocessing):
            self.preprocessing = preprocessing
        elif preprocessing == {}:
            self.preprocessing = preprocessing
        else:
            self.preprocessing = Preprocessing(**preprocessing)
        self.models = models


class ListModelsParameters(ExperimentConfig):

    config = {
        'models_parameters': 'models_params'
    }

    def __init__(self, models_parameters: Union[List[ModelsParameters], List[Dict], None] = None):

        if models_parameters is None:

            models_parameters_1 = ModelsParameters(ModelEmbedding.TFIDF,
                                                   Preprocessing(),
                                                   [TextSimilarityModels.BruteForce,
                                                    TextSimilarityModels.ClusterPruning])
            models_parameters_2 = ModelsParameters(ModelEmbedding.Transformer,
                                                   Preprocessing(),
                                                   [TextSimilarityModels.BruteForce])
            models_parameters_3 = ModelsParameters(ModelEmbedding.TransformerFineTuned,
                                                   Preprocessing(),
                                                   [TextSimilarityModels.BruteForce])
            models_parameters = [models_parameters_1, models_parameters_2, models_parameters_3]
        self.models_parameters = []
        for element in models_parameters:
            if isinstance(element, ModelsParameters):
                self.models_parameters.append(element)
            else:
                self.models_parameters.append(ModelsParameters(**element))


class DescriptionsColumnConfig(ExperimentConfig):
    """ Description Column configuration for starting an experiment: this object defines
    the role of specific columns in the dataset.

    Args:
        content_column (str, required): Name of the column containing the text descriptions in the
            description dataset.
        id_column (str, optional): Name of the id column in the description dataset.
    """

    config = {
    }

    def __init__(self,
                 content_column,
                 id_column):
        self.content_column = content_column
        self.id_column = id_column


class QueriesColumnConfig(ExperimentConfig):
    """ Description Column configuration for starting an experiment: this object defines
    the role of specific columns in the dataset.

    Args:
        content_column (str, required): Name of the column containing the text queries in the
            description dataset.
        id_column (str, optional): Name of the id column in the description dataset.
    """

    # config = {
    #     'queries_dataset_content_column': 'queriesDatasetContentColumn',
    #     'queries_dataset_matching_id_description_column': 'queriesDatasetMatchingIdDescriptionColumn',
    #     'queries_dataset_id_column': 'queriesDatasetIdColumn',
    # }

    def __init__(self,
                 queries_dataset_content_column,
                 queries_dataset_matching_id_description_column,
                 queries_dataset_id_column=None):
        self.queries_dataset_content_column = queries_dataset_content_column
        self.queries_dataset_matching_id_description_column = queries_dataset_matching_id_description_column
        self.queries_dataset_id_column = queries_dataset_id_column


class TextSimilarity(BaseExperimentVersion):
    """ A text similarity experiment version """

    default_metric: pio.metrics.TextSimilarity = pio.metrics.TextSimilarity.accuracy_at_k
    default_top_k: int = 10
    data_type: DataType = DataType.Tabular
    training_type: TypeProblem = TypeProblem.TextSimilarity
    resource: str = 'experiment-versions'
    model_class = TextSimilarityModel

    def __init__(self, **experiment_version_info):
        super().__init__(**experiment_version_info)

        self.predictions = {}
        self.predict_token = None

    def _update_from_dict(self, **experiment_version_info):
        super()._update_from_dict(**experiment_version_info)

        experiment_version_params = experiment_version_info['experiment_version_params']

        self.dataset_id: str = experiment_version_info['dataset_id']
        self.description_column_config = DescriptionsColumnConfig(
            content_column=experiment_version_params.get('content_column'),
            id_column=experiment_version_params.get('id_column'))
        self.metric: pio.metrics.TextSimilarity = pio.metrics.TextSimilarity(
            experiment_version_params.get('metric', self.default_metric))
        self.top_k: int = experiment_version_params.get('top_K', self.default_top_k)
        self.lang: TextSimilarityLang = TextSimilarityLang(experiment_version_params.get('lang',
                                                                                         TextSimilarityLang.Auto))

        if 'queries_dataset_id' in experiment_version_info:
            self.queries_dataset_id: str = experiment_version_info['queries_dataset_id']
            content_column = experiment_version_params.get('queries_dataset_content_column')
            matching_id = experiment_version_params.get('queries_dataset_matching_id_description_column')
            queries_dataset_id_column = experiment_version_params.get('queries_dataset_id_column', None)
            self.queries_column_config = QueriesColumnConfig(queries_dataset_content_column=content_column,
                                                             queries_dataset_matching_id_description_column=matching_id,
                                                             queries_dataset_id_column=queries_dataset_id_column)
        else:
            self.queries_dataset_id = None
            self.queries_column_config = None

        models_parameters = experiment_version_params.get('models_params')
        self.models_parameters = ListModelsParameters(models_parameters=models_parameters)

    @property
    def dataset(self) -> Dataset:
        """ Get the :class:`.Dataset` object corresponding to the training dataset of this experiment version.

        Returns:
            :class:`.Dataset`: Associated training dataset
        """
        return Dataset.from_id(self.dataset_id)

    @property
    def queries_dataset(self) -> Union[Dataset, None]:
        """ Get the :class:`.Dataset` object corresponding to the queries dataset of this experiment version.

        Returns:
            :class:`.Dataset`: Associated queries dataset
        """
        return Dataset.from_id(self.queries_dataset_id) if self.queries_dataset_id is not None else None

    @classmethod
    def from_id(cls, _id: str) -> 'TextSimilarity':
        """Get a text-similarity experiment version from the platform by its unique id.

        Args:
            _id (str): Unique id of the experiment version to retrieve

        Returns:
            :class:`.TextSimilarity`: Fetched experiment version

        Raises:
            PrevisionException: Any error while fetching data from the platform or parsing result
        """
        return cls(**super()._from_id(_id))

    @staticmethod
    def _build_experiment_version_creation_data(description, dataset, description_column_config, metric,
                                                top_k, lang, queries_dataset, queries_column_config,
                                                models_parameters,
                                                parent_version=None) -> Dict:
        data = super(TextSimilarity, TextSimilarity)._build_experiment_version_creation_data(
            description,
            parent_version=parent_version,
        )

        data['dataset_id'] = dataset.id
        data.update(to_json(description_column_config))
        data['metric'] = metric if isinstance(metric, str) else metric.value
        data['top_k'] = top_k
        data['lang'] = lang if isinstance(lang, str) else lang.value
        if queries_dataset is not None:
            if queries_column_config is None:
                raise ValueError('arg queries_column_config must be set if queries_dataset is set')
            data['queries_dataset_id'] = queries_dataset.id
            data.update(to_json(queries_column_config))
        data.update(to_json(models_parameters))

        return data

    @classmethod
    def _fit(
        cls,
        experiment_id: str,
        dataset: Dataset,
        description_column_config: DescriptionsColumnConfig,
        metric: pio.metrics.TextSimilarity = pio.metrics.TextSimilarity.accuracy_at_k,
        top_k: int = 10,
        lang: TextSimilarityLang = TextSimilarityLang.Auto,
        queries_dataset: Dataset = None,
        queries_column_config: Union[QueriesColumnConfig, None] = None,
        models_parameters: ListModelsParameters = ListModelsParameters(),
        description: str = None,
        parent_version: str = None,
    ) -> 'TextSimilarity':
        """ Start a text-similarity experiment version training with a specific configuration
        (on the platform).

        Args:
            experiment_id (str): The id of the experiment from which this version is created
            dataset (:class:`.Dataset`): Reference to the dataset object to use for as training dataset
            description_column_config (:class:`.DescriptionsColumnConfig`): Description column configuration
                (see the documentation of the :class:`.DescriptionsColumnConfig` resource for more details
                on each possible column types)
            metric (:class:`.metrics.TextSimilarity`, optional): Specific metric to use for the experiment
                (default: ``metrics.TextSimilarity.accuracy_at_k``)
            top_k (int, optional): top_k (default: ``10``)
            lang (:class:`.TextSimilarityLang`, optional): lang of the training dataset
                (default: ``.TextSimilarityLang.Auto``)
            queries_dataset (:class:`.Dataset`, optional): Reference to a dataset object to
                use as a queries dataset (default: ``None``)
            queries_column_config (:class:`.QueriesColumnConfig`): Queries column configuration
                (see the documentation of the :class:`.QueriesColumnConfig` resource for more details
                on each possible column types)
            models_parameters (:class:`.ListModelsParameters`): Specific training configuration
                (see the documentation of the :class:`.ListModelsParameters` resource for more details
                on all the parameters)
            description (str, optional): The description of this experiment version (default: ``None``)
            parent_version (str, optional): The parent version of this experiment_version (default: ``None``)

        Returns:
            :class:`.TextSimilarity`: Newly created text-similarity experiment version object
        """

        return super()._fit(
            experiment_id,
            description=description,
            parent_version=parent_version,
            dataset=dataset,
            description_column_config=description_column_config,
            metric=metric,
            top_k=top_k,
            lang=lang,
            queries_dataset=queries_dataset,
            queries_column_config=queries_column_config,
            models_parameters=models_parameters,
        )

    def new_version(
        self,
        dataset: Dataset = None,
        description_column_config: DescriptionsColumnConfig = None,
        metric: pio.metrics.TextSimilarity = None,
        top_k: int = None,
        lang: TextSimilarityLang = None,
        queries_dataset: Dataset = None,
        queries_column_config: Union[QueriesColumnConfig, None] = None,
        models_parameters: ListModelsParameters = None,
        description: str = None,
    ) -> 'TextSimilarity':
        """
        Start a new text-similarity experiment version training from this version (on the platform).
        The training parameters are copied from the current version and then overridden for those provided.

        Args:
            dataset (:class:`.Dataset`): Reference to the dataset object to use for as training dataset
            description_column_config (:class:`.DescriptionsColumnConfig`): Description column configuration
                (see the documentation of the :class:`.DescriptionsColumnConfig` resource for more details
                on each possible column types)
            metric (:class:`.metrics.TextSimilarity`, optional): Specific metric to use for the experiment
            top_k (int, optional): top_k
            lang (:class:`.TextSimilarityLang`, optional): lang of the training dataset
            queries_dataset (:class:`.Dataset`, optional): Reference to a dataset object to
                use as a queries dataset
            queries_column_config (:class:`.QueriesColumnConfig`): Queries column configuration
                (see the documentation of the :class:`.QueriesColumnConfig` resource for more details
                on each possible column types)
            models_parameters (:class:`.ListModelsParameters`): Specific training configuration
                (see the documentation of the :class:`.ListModelsParameters` resource for more details
                on all the parameters)
            description (str, optional): The description of this experiment version (default: ``None``)
        Returns:
            :class:`.TextSimilarity`: Newly created text-similarity experiment version object (new version)
        """
        return TextSimilarity._fit(
            self.experiment_id,
            dataset if dataset is not None else self.dataset,
            description_column_config if description_column_config is not None else self.description_column_config,
            metric=metric if metric is not None else self.metric,
            top_k=top_k if top_k is not None else self.top_k,
            lang=lang if lang is not None else self.lang,
            queries_dataset=queries_dataset if queries_dataset is not None else self.queries_dataset,
            queries_column_config=queries_column_config if queries_column_config is not None else
            self.queries_column_config,
            models_parameters=models_parameters if models_parameters is not None else self.models_parameters,
            description=description,
            parent_version=self.version,
        )



# -*- coding: utf-8 -*-
import json
from previsionio.model import Model
from typing import Dict, List, Union
import pandas as pd
import requests
import time
import datetime
import previsionio as pio
from functools import lru_cache
from dateutil import parser

from . import config
from .experiment_config import (AdvancedModel, DataType, Feature, NormalModel, Profile, SimpleModel,
                                TrainingConfig, ColumnConfig, TypeProblem, ExperimentState)
from .logger import logger
from .prevision_client import client
from .utils import parse_json, EventTuple, PrevisionException, zip_to_pandas, get_all_results
from .api_resource import ApiResource
from .dataset import Dataset


class BaseExperimentVersion(ApiResource):

    """Base parent class for all experiments objects."""

    training_config: TrainingConfig
    column_config: ColumnConfig
    id_key = 'experiment_id'

    resource = 'experiment-versions'
    training_type: TypeProblem
    data_type: DataType
    model_class: Model

    def __init__(self, **experiment_version_info):
        super().__init__(experiment_version_info['_id'])
        self._update_from_dict(**experiment_version_info)

        self._models = {}

    def _update_from_dict(self, **experiment_version_info):
        # we keep the raw entire dict, atm used only in print_info...
        self._experiment_version_info = experiment_version_info

        self.project_id: str = experiment_version_info.get('project_id')
        self.experiment_id: str = experiment_version_info.get('experiment_id')
        self.description: str = experiment_version_info.get('description')
        self.version = experiment_version_info['version']
        self.parent_version: str = experiment_version_info.get('parent_version')
        if 'experiment' in experiment_version_info and 'data_type' in experiment_version_info['experiment']:
            self.data_type: DataType = DataType(experiment_version_info['experiment']['data_type'])
        else:
            self.data_type = None
        if 'experiment' in experiment_version_info and 'training_type' in experiment_version_info['experiment']:
            self.training_type: TypeProblem = TypeProblem(experiment_version_info['experiment']['training_type'])
        else:
            self.training_type = None

        self.created_at: datetime.datetime = parser.parse(experiment_version_info.get('created_at'))

    def print_info(self):
        """ Print all info on the experiment. """
        # NOTE: maybe not set self._experiment_version_info and print each object attribut
        for k, v in self._experiment_version_info.items():
            print(str(k) + ': ' + str(v))

    # NOTE: this method is just here to parse raw_data (objects) and build the corresponding data (strings)
    #       that can be sent directly to the endpoint
    @staticmethod
    def _build_experiment_version_creation_data(description, parent_version=None) -> Dict:
        data = {
            'description': description,
            'parent_version': parent_version,
        }
        return data

    @classmethod
    def _new(cls, experiment_id, data) -> 'BaseExperimentVersion':
        endpoint = f'/experiments/{experiment_id}/versions'
        response = client.request(endpoint,
                                  method=requests.post,
                                  data=data,
                                  message_prefix='Experiment version creation')
        experiment_version_info = parse_json(response)
        experiment_version = cls(**experiment_version_info)
        return experiment_version

    def _update_draft(self, **kwargs):
        return self

    def _confirm(self) -> 'BaseExperimentVersion':
        endpoint = f'/experiment-versions/{self._id}/confirm'
        response = client.request(endpoint,
                                  method=requests.put,
                                  message_prefix='Experiment version confirmation')
        experiment_version_info = parse_json(response)
        self._update_from_dict(**experiment_version_info)
        return self

    @classmethod
    def _fit(
        cls,
        experiment_id: str,
        description: str = None,
        parent_version: str = None,
        **kwargs,
    ) -> 'BaseExperimentVersion':

        experiment_version_creation_data = cls._build_experiment_version_creation_data(
            description,
            parent_version=parent_version,
            **kwargs,
        )
        experiment_version_draft = cls._new(experiment_id, experiment_version_creation_data)
        experiment_version_draft._update_draft(**kwargs)
        experiment_version = experiment_version_draft._confirm()

        # NOTE: maybe update like that to be sure to have all the correct info of the resource
        # experiment_version._update_from_dict(**cls._from_id(experiment_version._id))

        return experiment_version

    def new_version(self, **kwargs):
        raise NotImplementedError

    def __len__(self):
        return len(self.models)

    def update_status(self):
        return super().update_status(specific_url='/{}/{}'.format(self.resource,
                                                                  self._id))

    @classmethod
    def _from_id(cls, _id: str) -> Dict:
        """Get an experiment from the platform by its unique id.

        Args:
            _id (str): Unique id of the experiment to retrieve
            version (int, optional): Specific version of the experiment to retrieve
                (default: 1)

        Returns:
            :class:`.BaseExperimentVersion`: Fetched experiment

        Raises:
            PrevisionException: Any error while fetching data from the platform
                or parsing result
        """
        return super()._from_id(specific_url='/{}/{}'.format(cls.resource, _id))

    # @property
    # def experiment(self) -> 'Experiment':
    #     """Get an experiment of current experiment version.

    #     Returns:
    #         :class:`.Experiment`: Fetched experiment

    #     Raises:
    #         PrevisionException: Any error while fetching data from the platform
    #             or parsing result
    #     """
    #     return Experiment.from_id(self.experiment_id)

    @property
    def models(self):
        """Get the list of models generated for the current experiment version. Only the models that
        are done training are retrieved.

        Returns:
            list(:class:`.Model`): List of models found by the platform for the experiment
        """
        end_point = '/{}/{}/models'.format(self.resource, self._id)
        models = get_all_results(client, end_point, method=requests.get)
        for model in models:
            if model['_id'] not in self._models:
                self._models[model['_id']] = self.model_class(**model)
        return list(self._models.values())

    @property
    @lru_cache()
    def train_dataset(self):
        """ Get the :class:`.Dataset` object corresponding to the training dataset
        of the experiment.

        Returns:
            :class:`.Dataset`: Associated training dataset
        """
        return self.dataset

    @property
    @lru_cache()
    def schema(self) -> dict:
        """ Get the data schema of the experiment.

        Returns:
            dict: Experiment schema
        """
        end_point = '/{}/{}/graph'.format(self.resource, self._id)
        response = client.request(endpoint=end_point,
                                  method=requests.get,
                                  message_prefix='Experiment schema')
        schema = json.loads(response.content.decode('utf-8'))
        return schema

    @property
    def best_model(self):
        """ Get the model with the best predictive performance over all models (including
        Blend models), where the best performance corresponds to a minimal loss.

        Returns:
            (:class:`.Model`, None): Model with the best performance in the experiment, or
            ``None`` if no model matched the search filter.
        """
        best_model = None
        if len(self.models) == 0:
            raise PrevisionException('models not ready yet')
        for model in self.models:
            if 'best' in model.tags:
                best_model = model

        if best_model is None:
            best_model = self.models[0]
        return best_model

    @property
    def fastest_model(self):
        """Returns the model that predicts with the lowest response time

        Returns:
            Model object -- corresponding to the fastest model
        """
        fastest_model = None
        if len(self.models) == 0:
            raise PrevisionException('models not ready yet')
        for model in self.models:
            if 'fastest' in model.tags:
                fastest_model = model

        if fastest_model is None:
            fastest_model = self.models[0]

        return fastest_model

    @property
    def done(self) -> bool:
        """ Get a flag indicating whether or not the experiment is currently done.

        Returns:
            bool: done status
        """
        status = self._status
        return status['state'] == ExperimentState.Done.value

    @property
    def running(self) -> bool:
        """ Get a flag indicating whether or not the experiment is currently running.

        Returns:
            bool: Running status
        """
        status = self._status
        return status['state'] == ExperimentState.Running.value

    @property
    def status(self) -> ExperimentState:
        """ Get a flag indicating whether or not the experiment is currently running.

        Returns:
            bool: Running status
        """
        status = self._status
        return ExperimentState(status['state'])

    def stop(self):
        """ Stop an experiment (stopping all nodes currently in progress). """
        logger.info('[Experiment] stopping experiment')
        end_point = '/{}/{}/stop'.format(self.resource, self._id)
        response = client.request(end_point,
                                  requests.put,
                                  message_prefix='Experiment stop')
        events_url = '/{}/{}'.format(self.resource, self._id)
        assert pio.client.event_manager is not None
        pio.client.event_manager.wait_for_event(self.resource_id,
                                                self.resource,
                                                EventTuple('EXPERIMENT_VERSION_UPDATE'),
                                                specific_url=events_url)
        logger.info('[Experiment] stopping:' + '  '.join(str(k) + ': ' + str(v)
                                                         for k, v in parse_json(response).items()))

    def delete(self):
        """Delete an experiment version from the actual [client] workspace.

        Raises:
            PrevisionException: If the experiment version does not exist
            requests.exceptions.ConnectionError: Error processing the request
        """
        super().delete()

    def wait_until(self, condition, raise_on_error: bool = True, timeout: float = config.default_timeout):
        """ Wait until condition is fulfilled, then break.

        Args:
            condition (func: (:class:`.BaseExperimentVersion`) -> bool.): Function to use to check the
                break condition
            raise_on_error (bool, optional): If true then the function will stop on error,
                otherwise it will continue waiting (default: ``True``)
            timeout (float, optional): Maximal amount of time to wait before forcing exit

        Example::

            experiment.wait_until(lambda experimentv: len(experimentv.models) > 3)

        Raises:
            PrevisionException: If the resource could not be fetched or there was a timeout.
        """
        t0 = time.time()
        while True:
            if timeout is not None and time.time() - t0 > timeout:
                raise PrevisionException('timeout while waiting on {}'.format(condition))

            try:
                if condition(self):
                    break
                elif self._status['state'] == 'failed':
                    raise PrevisionException('Resource failed while waiting')
            except PrevisionException as e:
                logger.warning(e.__repr__())
                if raise_on_error:
                    raise

            time.sleep(config.scheduler_refresh_rate)

    def get_holdout_predictions(self, full: bool = False):
        """
        Retrieves the list of holdout predictions for the current experiment from client workspace
        (with the full predictions object if necessary)
        Args:
            full (boolean): If true, return full holdout prediction objects (else only metadata)
        """
        end_point = '/experiment-versions/{}/holdout-predictions'.format(self._id)
        response = client.request(endpoint=end_point,
                                  method=requests.get,
                                  message_prefix='Holdout predictions listing')
        preds_list = (json.loads(response.content.decode('utf-8')))['items']
        preds_dict = {}
        for pred in preds_list:
            _id = pred.pop('_id')
            if full:
                end_point = '/predictions/{}/download'.format(_id)
                response = client.request(endpoint=end_point,
                                          method=requests.get,
                                          message_prefix='Predictions fetching')
                preds_dict[_id] = zip_to_pandas(response)
            else:
                preds_dict[_id] = pred
        return preds_dict

    def get_predictions(self, full: bool = False):
        """
        Retrieves the list of predictions for the current experiment from client workspace
        (with the full predictions object if necessary)
        Args:
            full (boolean): If true, return full prediction objects (else only metadata)
        """
        response = client.request(endpoint='/experiments/{}/predictions'.format(self._id),
                                  method=requests.get,
                                  message_prefix='Predictions listing')
        preds_list = (json.loads(response.content.decode('utf-8')))['items']
        preds_dict = {}
        for pred in preds_list:
            _id = pred.pop('_id')
            if full:
                response = client.request(endpoint='/predictions/{}/download'.format(_id),
                                          method=requests.get,
                                          message_prefix='Predictions fetching')
                preds_dict[_id] = zip_to_pandas(response)
            else:
                preds_dict[_id] = pred
        return preds_dict

    def delete_prediction(self, prediction_id: str):
        """ Delete a prediction in the list for the current experiment from the actual [client] workspace.

        Args:
            prediction_id (str): Unique id of the prediction to delete

        Returns:
            dict: Deletion process results
        """
        endpoint = '/experiments/{}/versions/{}/predictions/{}'.format(self._id, self.version, prediction_id)
        response = client.request(endpoint=endpoint,
                                  method=requests.delete,
                                  message_prefix='Prediction delete')
        return (json.loads(response.content.decode('utf-8')))

    def delete_predictions(self):
        """ Delete all predictions in the list for the current experiment from the actual [client] workspace.

        Returns:
            dict: Deletion process results
        """
        response = client.request(endpoint='/experiments/{}/versions/{}/predictions'.format(self._id, self.version),
                                  method=requests.delete,
                                  message_prefix='Predictions delete')
        return (json.loads(response.content.decode('utf-8')))

    @property
    def score(self) -> float:
        """ Get the current score of the experiment (i.e. the score of the model that is
        currently considered the best performance-wise for this experiment).

        Returns:
            float: Experiment score (or infinity if not available).
        """
        try:
            return self._status['best_model']['metric']['value']
        except KeyError:
            PrevisionException('Score not available yet')


class ClassicExperimentVersion(BaseExperimentVersion):

    def __init__(self, **experiment_version_info):
        super().__init__(**experiment_version_info)

        self.predictions = {}
        self.predict_token = None

    def _update_from_dict(self, **experiment_version_info):
        super()._update_from_dict(**experiment_version_info)

        self.dataset_id: str = experiment_version_info['dataset_id']

        experiment_params = experiment_version_info['experiment_version_params']
        self.column_config = ColumnConfig(target_column=experiment_params.get('target_column'),
                                          fold_column=experiment_params.get('fold_column'),
                                          id_column=experiment_params.get('id_column'),
                                          weight_column=experiment_params.get('weight_column'),
                                          time_column=experiment_params.get('time_column', None),
                                          group_columns=experiment_params.get('group_columns'),
                                          apriori_columns=experiment_params.get('apriori_columns'),
                                          drop_list=experiment_params.get('drop_list'))

        self.metric: str = experiment_params['metric']

        self.holdout_dataset_id: Union[str, None] = experiment_version_info.get('holdout_dataset_id', None)

        self.training_config = TrainingConfig(
            profile=Profile(experiment_params.get('profile')),
            features=[Feature(f) for f in experiment_params.get('features_engineering_selected_list', [])],
            advanced_models=[AdvancedModel(f) for f in experiment_params.get('normal_models', [])],
            normal_models=[NormalModel(f) for f in experiment_params.get('lite_models', [])],
            simple_models=[SimpleModel(f) for f in experiment_params.get('simple_models', [])],
            feature_time_seconds=experiment_params.get('features_selection_time', 3600),
            feature_number_kept=experiment_params.get('features_selection_count', None))

    @property
    def dataset(self) -> Dataset:
        """ Get the :class:`.Dataset` object corresponding to the training dataset of this experiment version.

        Returns:
            :class:`.Dataset`: Associated training dataset
        """
        return Dataset.from_id(self.dataset_id)

    @property
    def holdout_dataset(self) -> Union[Dataset, None]:
        """ Get the :class:`.Dataset` object corresponding to the holdout dataset of this experiment version.

        Returns:
            :class:`.Dataset`: Associated holdout dataset
        """
        return Dataset.from_id(self.holdout_dataset_id) if self.holdout_dataset_id is not None else None

    @property
    def advanced_models_list(self) -> List[AdvancedModel]:
        """ Get the list of selected advanced models in the experiment.

        Returns:
            list(AdvancedModel): Names of the normal models selected for the experiment
        """
        return [AdvancedModel(f) for f in self._status['experiment_version_params'].get('normal_models', [])]

    @property
    def normal_models_list(self) -> List[NormalModel]:
        """ Get the list of selected normal models in the experiment.

        Returns:
            list(NormalModel): Names of the normal models selected for the experiment
        """
        return [NormalModel(f) for f in self._status['experiment_version_params'].get('lite_models', [])]

    @property
    def simple_models_list(self) -> List[SimpleModel]:
        """ Get the list of selected simple models in the experiment.

        Returns:
            list(SimpleModel): Names of the simple models selected for the experiment
        """
        return [SimpleModel(f) for f in self._status['experiment_version_params'].get('simple_models', [])]

    @property
    @lru_cache()
    def correlation_matrix(self) -> pd.DataFrame:
        """ Get the correlation matrix of the features (those constitute the dataset
        on which the experiment was trained).

        Returns:
            ``pd.DataFrame``: Correlation matrix as a ``pandas`` dataframe
        """
        end_point = '/{}/{}/correlation-matrix'.format(self.resource, self._id)
        response = client.request(endpoint=end_point,
                                  method=requests.get,
                                  message_prefix='Correlation matrix fetching')
        corr = json.loads(response.content.decode('utf-8'))
        var_names = [d['name'] for d in corr]
        matrix = pd.DataFrame(0, index=var_names, columns=var_names)
        for i, name in enumerate(var_names):
            matrix[name] = [d['score'] for d in corr[i]['correlation']]
        return matrix

    @property
    @lru_cache()
    def features(self) -> List[dict]:
        """ Get the general description of the experiment's features, such as:

        - feature types distribution
        - feature information list
        - list of dropped features

        Returns:
            dict: General features information
        """
        # todo pagination
        url = '/{}/{}/features'.format(self.resource, self._id)
        return get_all_results(client, url, method=requests.get)

    @property
    @lru_cache()
    def features_stats(self):
        """ Get the general description of the experiment's features, such as:

        - feature types distribution
        - feature information list
        - list of dropped features

        Returns:
            dict: General features information
        """
        # todo pagination
        end_point = '/{}/{}/features-stats'.format(self.resource, self._id)
        response = client.request(endpoint=end_point,
                                  method=requests.get,
                                  message_prefix='Features stats fetching')
        return (json.loads(response.content.decode('utf-8')))

    @property
    @lru_cache()
    def dropped_features(self):
        """ Get dropped features

        Returns:
            dict: Dropped features
        """
        # todo pagination
        end_point = '/{}/{}/dropped-features'.format(self.resource, self._id)
        response = client.request(endpoint=end_point,
                                  method=requests.get,
                                  message_prefix='Dropped features fetching')
        return (json.loads(response.content.decode('utf-8')))

    def get_feature_info(self, feature_name: str) -> Dict:
        """ Return some information about the given feature, such as:

        - name: the name of the feature as it was given in the ``feature_name`` parameter
        - type: linear, categorical, ordinal...
        - stats: some basic statistics such as number of missing values, (non missing) values
          count, plus additional information depending on the feature type:

            - for a linear feature: min, max, mean, std and median
            - for a categorical/textual feature: modalities/words frequencies,
              list of the most frequent tokens

        - role: whether or not the feature is a target/fold/weight or id feature (and for time
          series experiments, whether or not it is a group/apriori feature - check the
          `Prevision.io's timeseries documentation
          <https://previsionio.readthedocs.io/fr/latest/projects_timeseries.html>`_)
        - importance_value: scores reflecting the importance of the given feature

        Args:
            feature_name (str): Name of the feature to get informations about

            .. warning::

                The ``feature_name`` is case-sensitive, so "age" and "Age" are different
                features!

        Returns:
            dict: Dictionary containing the feature information

        Raises:
            PrevisionException: If the given feature name does not match any feaure
        """
        endpoint = '/{}/{}/features/{}'.format(self.resource, self._id, feature_name)
        response = client.request(endpoint=endpoint,
                                  method=requests.get,
                                  message_prefix='Features info fetching')

        result = (json.loads(response.content.decode('utf-8')))

        # drop chart-related informations
        keep_list = list(filter(lambda x: 'chart' not in x.lower(),
                                result.keys())
                         )
        return {k: result[k] for k in keep_list}

    @property
    def drop_list(self) -> List[str]:
        """ Get the list of drop columns in the experiment.

        Returns:
            list(str): Names of the columns dropped from the dataset
        """
        return self._status['experiment_version_params'].get('drop_list', [])

    @property
    def feature_list(self) -> List[Feature]:
        """ Get the list of selected feature engineering modules in the experiment.

        Returns:
            list(str): Names of the feature engineering modules selected for the experiment
        """
        res = [Feature(f) for f in self._status['experiment_version_params'].get(
            'features_engineering_selected_list', [])]
        return res

    def get_cv(self) -> pd.DataFrame:
        """ Get the cross validation dataset from the best model of the experiment.

        Returns:
            ``pd.DataFrame``: Cross validation dataset
        """

        best = self.best_model

        return best.cross_validation

    def predict_from_dataset(self,
                             dataset,
                             confidence=False,
                             dataset_folder=None) -> pd.DataFrame:
        """ Get the predictions for a dataset stored in the current active [client]
        workspace using the best model of the experiment.

        Arguments:
            dataset (:class:`.Dataset`): Reference to the dataset object to make
                predictions for
            confidence (bool, optional): Whether to predict with confidence values
                (default: ``False``)
            dataset_folder (:class:`.Dataset`): Matching folder dataset for the
                predictions, if necessary

        Returns:
            :class:`previsionio.prediction.ValidationPrediction`: The registered prediction object in the current
            workspace
        """

        best = self.best_model

        return best.predict_from_dataset(dataset, confidence=confidence, dataset_folder=dataset_folder)

    def predict(self, df, confidence=False, prediction_dataset_name=None) -> pd.DataFrame:
        """ Get the predictions for a dataset stored in the current active [client]
        workspace using the best model of the experiment with a Scikit-learn style blocking prediction mode.

        .. warning::

            For large dataframes and complex (blend) models, this can be slow (up to 1-2 hours).
            Prefer using this for simple models and small dataframes, or use option ``use_best_single = True``.

        Args:
            df (``pd.DataFrame``): ``pandas`` DataFrame containing the test data
            confidence (bool, optional): Whether to predict with confidence values
                (default: ``False``)

        Returns:
            ``pd.DataFrame``: Prediction results dataframe
        """

        best = self.best_model

        return best.predict(df=df, confidence=confidence, prediction_dataset_name=prediction_dataset_name)



from enum import Enum
from typing import Dict
import requests
from . import client
from .utils import parse_json
from .api_resource import ApiResource, UniqueResourceMixin


class GCloud(Enum):
    """
    Google services."""
    big_query = 'BigQuery'
    """Google BigQuery"""
    storage = 'Storage'
    """Google Storage"""


class Connector(ApiResource, UniqueResourceMixin):
    """ A connector to interact with a distant source of data (and
    easily get data snapshots using an associated :class:`.DataSource`
    resource).

    Args:
        _id (str): Unique reference of the connector on the platform
        project_id (str): Unique reference of the project id on the platform
        name (str): Name of the connector
    """

    resource = 'connectors'
    conn_type = 'connector'

    def __init__(self, _id: str, project_id: str, name: str):
        super().__init__(_id=_id)
        self.project_id = project_id
        self.name = name

    @classmethod
    def list(cls, project_id: str, all: bool = False):
        """ List all the available connectors in the current active [client] workspace.

        .. warning::

            Contrary to the parent ``list()`` function, this method
            returns actual :class:`.Connector` objects rather than
            plain dictionaries with the corresponding data.

        Args:
            project_id (str): Unique reference of the project id on the platform
            all (boolean, optional): Whether to force the SDK to load all items of
                the given type (by calling the paginated API several times). Else,
                the query will only return the first page of result.

        Returns:
            list(:class:`.Connector`): Fetched connector objects
        """
        resources = super()._list(all=all, project_id=project_id)
        return [connectors_names.get(conn_data['type']).from_dict(conn_data) for conn_data in resources
                if conn_data['type'] == cls.conn_type or cls.conn_type == 'connector']

    @classmethod
    def _create_connector(cls, project_id: str, data: Dict, content_type: str = None):
        data['type'] = cls.conn_type
        resp = client.request('/projects/{}/{}'.format(project_id, cls.resource),
                              data=data,
                              method=requests.post,
                              message_prefix='New connector creation',
                              content_type=content_type)

        connector_info = parse_json(resp)
        if '_id' not in connector_info:
            if 'message' in connector_info:
                raise Exception('Prevision.io error: {}'.format(' '.join(connector_info['message'])))
            else:
                raise Exception('unknown error:{}'.format(connector_info))

        return connector_info

    @classmethod
    def _new(cls, project_id: str, name: str, host: str, port: int, username: str, password: str):
        """ Create a new connector object on the platform.

        Args:
            project_id (str): Unique reference of the project id on the platform
            name (str): Name of the connector
            host (str): Url of the connector
            port (int): Port of the connector
            username (str): Username to use connect to the remote data source
            password (str): Password to use connect to the remote data source

        Returns:
            :class:`.Connector`: Newly create connector object
        """
        data = {
            'name': name,
            'host': host,
            'port': port,
            'username': username,
            'password': password,
        }
        connector_info = cls._create_connector(project_id, data)
        return cls.from_dict(connector_info)

    def test(self):
        """ Test a connector already uploaded on the platform.

        Returns:
            dict: Test results
        """
        resp = client.request('/connectors/{}/test'.format(self.id), method=requests.post, check_response=False)
        if resp.status_code == 200:
            return True
        else:
            return False

    def delete(self):
        """Delete a connector from the actual [client] workspace.

        Raises:
            PrevisionException: If the connector does not exist
            requests.exceptions.ConnectionError: Error processing the request
        """
        super().delete()


class DataTableBaseConnector(Connector):
    """ A specific type of connector to interact with a database client (containing databases and tables), and
    easily get data snapshots using an associated :class:`.DataSource`
    resource).

    Args:
        _id (str): Unique reference of the connector on the platform
        project_id (str): Unique reference of the project id on the platform
        name (str): Name of the connector
        host (str): Url of the connector
        port (int): Port of the connector
        username (str): Username to use connect to the remote data source
    """

    def __init__(self, _id: str, project_id: str, name: str, host: str, port: int, username: str):
        super().__init__(_id=_id, project_id=project_id, name=name)
        self.host = host
        self.port = port
        self.username = username

    @classmethod
    def from_dict(cls, connector_info: Dict):
        connector = cls(
            connector_info['_id'],
            connector_info['project_id'],
            connector_info['name'],
            connector_info['host'],
            connector_info['port'],
            connector_info['username'],
        )
        return connector

    def list_databases(self):
        """ List all available databases for the client.

        Returns:
            dict: Databases information
        """
        url = '/{}/{}/databases'.format(self.resource, self._id)
        resp = client.request(url, requests.get, message_prefix='Databases listing')
        resp_json = parse_json(resp)
        return resp_json['items']

    def list_tables(self, database: str):
        """ List all available tables in a specific database for the client.

        Args:
            database (str): Name of the database to find tables for

        Returns:
            dict: Tables information
        """
        url = '/{}/{}/databases/{}/tables'.format(self.resource, self._id, database)
        resp = client.request(url, requests.get, message_prefix='Tables listing')
        resp_json = parse_json(resp)
        return resp_json['items']


class DataFileBaseConnector(Connector):
    """ A specific type of connector to interact with a database client (containing files), and
    easily get data snapshots using an associated :class:`.DataSource`
    resource).

    Args:
        _id (str): Unique reference of the connector on the platform
        project_id (str): Unique reference of the project id on the platform
        name (str): Name of the connector
        host (str): Url of the connector
        port (int): Port of the connector
        username (str): Username to use connect to the remote data source
    """

    def __init__(self, _id: str, project_id: str, name: str, host: str, port: int, username: str):
        super().__init__(_id=_id, project_id=project_id, name=name)
        self.host = host
        self.port = port
        self.username = username

    @classmethod
    def from_dict(cls, connector_info: Dict):
        connector = cls(
            connector_info['_id'],
            connector_info['project_id'],
            connector_info['name'],
            connector_info['host'],
            connector_info['port'],
            connector_info['username'],
        )
        return connector

    def list_files(self):
        """ List all available tables in a specific database for the client.

        Args:
            database (str): Name of the database to find tables for

        Returns:
            dict: files information
        """
        url = '/{}/{}/paths'.format(self.resource, self._id)
        resp = client.request(url, requests.get, message_prefix='Datasource files listing')
        resp_json = parse_json(resp)
        return resp_json['items']


class FTPConnector(DataFileBaseConnector):
    """ A specific type of connector to interact with a FTP client (containing files), and
    easily get data snapshots using an associated :class:`.DataSource`
    resource).

    Args:
        _id (str): Unique reference of the connector on the platform
        project_id (str): Unique reference of the project id on the platform
        name (str): Name of the connector
        host (str): Url of the connector
        port (int): Port of the connector
        username (str): Username to use connect to the remote data source
    """

    conn_type = 'FTP'


class SFTPConnector(DataFileBaseConnector):
    """ A specific type of connector to interact with a secured FTP client (containing files), and
    easily get data snapshots using an associated :class:`.DataSource`
    resource).

    Args:
        _id (str): Unique reference of the connector on the platform
        project_id (str): Unique reference of the project id on the platform
        name (str): Name of the connector
        host (str): Url of the connector
        port (int): Port of the connector
        username (str): Username to use connect to the remote data source
    """

    conn_type = 'SFTP'


class SQLConnector(DataTableBaseConnector):
    """ A specific type of connector to interact with a SQL database client (containing databases and tables), and
    easily get data snapshots using an associated :class:`.DataSource`
    resource).

    Args:
        _id (str): Unique reference of the connector on the platform
        project_id (str): Unique reference of the project id on the platform
        name (str): Name of the connector
        host (str): Url of the connector
        port (int): Port of the connector
        username (str): Username to use connect to the remote data source
    """

    conn_type = 'SQL'


class S3Connector(Connector):
    """ A specific type of connector to interact with an Amazon S3 client (containing buckets with files),
    and easily get data snapshots using an associated :class:`.DataSource` resource).

    Args:
        _id (str): Unique reference of the connector on the platform
        project_id (str): Unique reference of the project id on the platform
        name (str): Name of the connector
        host (str): Url of the connector
        port (int): Port of the connector
        username (str): Username to use connect to the remote data source
    """

    conn_type = 'S3'

    def __init__(self, _id: str, project_id: str, name: str, username: str):
        super().__init__(_id=_id, project_id=project_id, name=name)
        self.username = username

    @classmethod
    def from_dict(cls, connector_info: Dict):
        connector = cls(
            connector_info['_id'],
            connector_info['project_id'],
            connector_info['name'],
            connector_info['username'],
        )
        return connector

    @classmethod
    def _new(cls, project_id: str, name: str, username: str, password: str):
        """ Create a new connector object on the platform.

        Args:
            project_id (str): Unique reference of the project id on the platform
            name (str): Name of the connector
            username (str): Username to use connect to the remote data source
            password (str): Password to use connect to the remote data source

        Returns:
            :class:`.S3Connector`: Newly create connector object
        """
        data = {
            'name': name,
            'username': username,
            'password': password,
        }
        connector_info = cls._create_connector(project_id, data)
        return cls.from_dict(connector_info)


class GCPConnector(Connector):
    """ A specific type of connector to interact with a GCP database client
    (containing databases and tables or buckets), and easily get data snapshots
    using an associated :class:`.DataSource` resource).

    Args:
        _id (str): Unique reference of the connector on the platform
        project_id (str): Unique reference of the project id on the platform
        name (str): Name of the connector
    """

    conn_type = 'GCP'

    @classmethod
    def from_dict(cls, connector_info: Dict):
        connector = cls(
            connector_info['_id'],
            connector_info['project_id'],
            connector_info['name'],
        )
        return connector

    @classmethod
    def _new(cls, project_id: str, name: str, googleCredentials: Dict):
        """ Create a new connector object on the platform.

        Args:
            project_id (str): Unique reference of the project id on the platform
            name (str): Name of the connector
            googleCredentials(dict): google credentials

        Returns:
            :class:`.GCPConnector`: Newly create connector object
        """
        data = {
            'name': name,
            'googleCredentials': googleCredentials,
        }
        connector_info = cls._create_connector(project_id, data, content_type='application/json')
        return cls.from_dict(connector_info)


connectors_names = {
    'SQL': SQLConnector,
    'FTP': FTPConnector,
    'SFTP': SFTPConnector,
    'S3': S3Connector,
    'GCP': GCPConnector,
}



from typing import Dict
from requests.models import Response
from requests_oauthlib import OAuth2Session
from oauthlib.oauth2 import BackendApplicationClient
from previsionio.utils import NpEncoder
import json
import time
import requests
from . import logger
from . import config
from .utils import handle_error_response, parse_json, PrevisionException
from .prevision_client import client


class DeployedModel(object):

    """
    DeployedModel class to interact with a deployed model.

    Args:
        prevision_app_url (str): URL of the App. Can be retrieved on your app dashbord.
        client_id (str): Your app client id. Can be retrieved on your app dashbord.
        client_secret (str): Your app client secret. Can be retrieved on your app dashbord.
        prevision_token_url (str): URL to get the OAuth2 token of the deployed model. Required
            only if working on-premise (custom IP address) otherwise it is retrieved automatically.
    """

    def __init__(
        self,
        prevision_app_url: str,
        client_id: str,
        client_secret: str,
        prevision_token_url: str = None,
    ):
        """Init DeployedModel (and check that the connection is valid)."""
        self.prevision_app_url = prevision_app_url
        self.client_id = client_id
        self.client_secret = client_secret

        if prevision_token_url:
            self.prevision_token_url = prevision_token_url
        else:
            try:
                version_resp = client.request(
                    '/version',
                    method=requests.get,
                    message_prefix='Getting deployed model oidc_url',
                )
                version_resp = parse_json(version_resp)
                self.prevision_token_url = version_resp['oidc_url']
                self.prevision_token_url += '/auth/realms/prevision.io/protocol/openid-connect/token'
            except Exception as e:
                logger.error(e)
                raise PrevisionException(f'Cannot get prevision_token_url: {e}')

        self.problem_type = None
        self.token = None

        try:
            about_resp = self.request('/about', method=requests.get)
            app_info = parse_json(about_resp)
            self.problem_type = app_info['problem_type']
            inputs_resp = self.request('/inputs', method=requests.get)
            self.inputs = parse_json(inputs_resp)
            outputs_resp = self.request('/outputs', method=requests.get)
            self.outputs = parse_json(outputs_resp)
        except Exception as e:
            logger.error(e)
            raise PrevisionException(f'Cannot connect: {e}')

    def _generate_token(self):
        client = BackendApplicationClient(client_id=self.client_id)
        oauth = OAuth2Session(client=client)
        token = oauth.fetch_token(token_url=self.prevision_token_url,
                                  client_id=self.client_id,
                                  client_secret=self.client_secret)
        self.token = token
        return token

    def _get_token(self):
        while self.token is None or time.time() > self.token['expires_at'] - 60:
            try:
                self._generate_token()
            except Exception as e:
                logger.warning(f'failed to generate token with error {e.__repr__()}')
                time.sleep(.5)

    def _check_token_url_app(self):

        if not self.prevision_app_url:
            raise PrevisionException('No url configured. Call client_app.init_client() to initialize')

        if not self.client_id:
            raise PrevisionException('No client id configured. Call client_app.init_client() to initialize')

        if not self.client_secret:
            raise PrevisionException('No client secret configured. Call client_app.init_client() to initialize')

    def request(
        self,
        endpoint: str,
        method,
        files: Dict = None,
        data: Dict = None,
        allow_redirects: bool = True,
        content_type: str = None,
        check_response: bool = True,
        message_prefix: str = None,
        **requests_kwargs,
    ):
        """
        Make a request on the desired endpoint with the specified method & data.

        Requires initialization.

        Args:
            endpoint: (str): api endpoint (e.g. /experiments, /prediction/file)
            method (requests.{get,post,delete}): requests method
            files (dict): files dict
            data (dict): for single predict
            content_type (str): force request content-type
            allow_redirects (bool): passed to requests method

        Returns:
            request response

        Raises:
            Exception: Error if url/token not configured
        """
        self._check_token_url_app()

        url = self.prevision_app_url + endpoint

        status_code = 502
        retries = config.request_retries
        n_tries = 0
        resp = None
        while (n_tries < retries) and (status_code in config.retry_codes):
            n_tries += 1
            try:
                self._get_token()
                assert self.token is not None
                headers = {
                    "Authorization": "Bearer " + self.token['access_token'],
                }
                if content_type:
                    headers['content-type'] = content_type

                resp = method(url,
                              headers=headers,
                              files=files,
                              allow_redirects=allow_redirects,
                              data=data,
                              **requests_kwargs)
                status_code = resp.status_code

            except Exception as e:
                raise PrevisionException(f'Error requesting: {url} with error {e.__repr__()}')

            if status_code in config.retry_codes:
                logger.warning(f'Failed to request {url} with status code {status_code}.'
                               f' Retrying {retries - n_tries} times')
                time.sleep(config.request_retry_time)

        assert isinstance(resp, Response)
        if check_response:
            handle_error_response(resp, url, data, message_prefix=message_prefix, n_tries=n_tries)

        return resp

    def predict(
        self,
        predict_data: Dict,
        use_confidence: bool = False,
        explain: bool = False,
    ):
        """ Get a prediction on a single instance using the best model of the experiment.

        Args:
            predict_data (dictionary): input data for prediction
            confidence (bool, optional): Whether to predict with confidence values
                (default: ``False``)
            explain (bool): Whether to explain prediction (default: ``False``)

        Returns:
            tuple(float, float, dict): Tuple containing the prediction value, confidence and explain.
            In case of regression problem type, confidence format is a list.
            In case of multiclassification problem type, prediction value format is a string.

        """

        predict_url = '/predict'

        if explain or use_confidence:
            predict_url += '?'

        if explain:
            predict_url += 'explain=true&'

        if use_confidence:
            predict_url += 'confidence=true'

        predict_url = predict_url.rstrip('&')
        resp = self.request(predict_url,
                            data=json.dumps(predict_data, cls=NpEncoder),
                            method=requests.post,
                            message_prefix='Deployed model predict')

        pred_response = resp.json()
        target_name = self.outputs[0]['name']
        preds = pred_response['response']['predictions']
        prediction = preds[target_name]
        if use_confidence:
            if self.problem_type == 'regression':
                confidance_resp = [{key: value} for key, value in preds.items() if 'TARGET_quantile=' in key]
            elif 'confidence' in preds:
                confidance_resp = preds['confidence']
            else:
                confidance_resp = None
        else:
            confidance_resp = None

        if explain and 'explanation' in preds:
            explain_resp = preds['explanation']
        else:
            explain_resp = None

        return prediction, confidance_resp, explain_resp



# -*- coding: utf-8 -*-
from typing import Dict, Tuple, Union

from .dataset import Dataset, DatasetImages
from .experiment_config import DataType, TypeProblem, TrainingConfig, ColumnConfig
from . import metrics
from .experiment_version import ClassicExperimentVersion
from .model import RegressionModel, ClassificationModel, MultiClassificationModel
from .utils import to_json

MODEL_CLASS_DICT = {
    TypeProblem.Regression: RegressionModel,
    TypeProblem.Classification: ClassificationModel,
    TypeProblem.MultiClassification: MultiClassificationModel
}


class Supervised(ClassicExperimentVersion):

    """ A supervised experiment version """

    def __init__(self, **experiment_version_info):
        super().__init__(**experiment_version_info)

    def _update_from_dict(self, **experiment_version_info):
        super()._update_from_dict(**experiment_version_info)
        if self.data_type == DataType.Images:
            self.dataset_images_id: str = experiment_version_info['folder_dataset_id']
        self.model_class = MODEL_CLASS_DICT.get(self.training_type, RegressionModel)

    @property
    def dataset_images(self) -> DatasetImages:
        """ Get the :class:`.DatasetImages` object corresponding to the images training dataset of this
        experiment version. Available only if data_type of this experiment_version is images

        Returns:
            :class:`.DatasetImages`: Associated images training dataset
        """
        if self.data_type != DataType.Images:
            raise RuntimeError(f'dataset_images not available when data type is not {DataType.Images.value}')
        return DatasetImages.from_id(self.dataset_images_id)

    @classmethod
    def from_id(cls, _id: str) -> 'Supervised':
        """Get a supervised experiment version from the platform by its unique id.

        Args:
            _id (str): Unique id of the experiment version to retrieve

        Returns:
            :class:`.Supervised`: Fetched experiment version

        Raises:
            PrevisionException: Any error while fetching data from the platform or parsing result
        """
        return cls(**super()._from_id(_id))

    @staticmethod
    def _build_experiment_version_creation_data(description, dataset, column_config, metric,
                                                holdout_dataset, training_config,
                                                parent_version=None) -> Dict:
        data = super(Supervised, Supervised)._build_experiment_version_creation_data(
            description,
            parent_version=parent_version,
        )

        # because if Image there is the dataset and the images zip
        if isinstance(dataset, tuple):
            dataset, folder_dataset = dataset
            data['dataset_id'] = dataset.id
            if folder_dataset is not None:
                data['folder_dataset_id'] = folder_dataset.id
        else:
            data['dataset_id'] = dataset.id

        data.update(to_json(column_config))
        data['metric'] = metric if isinstance(metric, str) else metric.value
        data['holdout_dataset_id'] = holdout_dataset.id if holdout_dataset is not None else None
        data.update(to_json(training_config))

        return data

    @classmethod
    def _fit(
        cls,
        experiment_id: str,
        dataset: Union[Dataset, Tuple[Dataset, DatasetImages]],
        column_config: ColumnConfig,
        metric: metrics.Enum,
        holdout_dataset: Dataset = None,
        training_config: TrainingConfig = TrainingConfig(),
        description: str = None,
        parent_version: str = None,
    ) -> 'Supervised':
        """ Start a supervised experiment version training with a specific configuration (on the platform).

        Args:
            experiment_id (str): The id of the experiment from which this version is created
            dataset (:class:`.Dataset`, :class:`.DatasetImages`): Reference to the dataset(s)
                object(s) to use for as training dataset(s)
            column_config (:class:`.ColumnConfig`): Column configuration for the experiment
                (see the documentation of the :class:`.ColumnConfig` resource for more details
                on each possible column types)
            metric (metrics.Enum): Specific metric to use for the experiment version
            holdout_dataset (:class:`.Dataset`, optional): Reference to a dataset object to
                use as a holdout dataset (default: ``None``)
            training_config (:class:`.TrainingConfig`, optional): Specific training configuration
                (see the documentation of the :class:`.TrainingConfig` resource for more details
                on all the parameters)
            description (str, optional): The description of this experiment version (default: ``None``)
            parent_version (str, optional): The parent version of this experiment_version (default: ``None``)


        Returns:
            :class:`.Supervised`: Newly created supervised experiment version object
        """
        return super()._fit(
            experiment_id,
            description=description,
            parent_version=parent_version,
            dataset=dataset,
            column_config=column_config,
            metric=metric,
            holdout_dataset=holdout_dataset,
            training_config=training_config,
        )

    def new_version(
        self,
        dataset: Dataset = None,
        dataset_images: DatasetImages = None,
        column_config: ColumnConfig = None,
        metric: metrics.Enum = None,
        holdout_dataset: Dataset = None,
        training_config: TrainingConfig = None,
        description: str = None,
    ) -> 'Supervised':
        """
        Start a new supervised experiment version training from this version (on the platform).
        The training parameters are copied from the current version and then overridden for those provided.

        Args:
            dataset (:class:`.Dataset`): Reference to the dataset
                object to use for as training dataset
            dataset_images (:class:`.DatasetImages`): Reference to the images dataset
                object to use for as training dataset
            column_config (:class:`.ColumnConfig`, optional): Column configuration for the experiment
                (see the documentation of the :class:`.ColumnConfig` resource for more details
                on each possible column types)
            metric (metrics.Enum, optional): Specific metric to use for the experiment version
            holdout_dataset (:class:`.Dataset`, optional): Reference to a dataset object to
                use as a holdout dataset (default: ``None``)
            training_config (:class:`.TrainingConfig`, optional): Specific training configuration
                (see the documentation of the :class:`.TrainingConfig` resource for more details
                on all the parameters)
            description (str, optional): The description of this experiment version (default: ``None``)
        Returns:
            :class:`.Supervised`: Newly created supervised experiment object (new version)
        """
        dataset = dataset if dataset is not None else self.dataset
        if self.data_type == DataType.Images:
            dataset_images = dataset_images if dataset_images is not None else self.dataset_images
            dataset = (dataset, dataset_images)
        return Supervised._fit(
            self.experiment_id,
            dataset,
            column_config if column_config is not None else self.column_config,
            metric if metric is not None else self.metric,
            holdout_dataset=holdout_dataset if holdout_dataset is not None else self.holdout_dataset,
            training_config=training_config if training_config is not None else self.training_config,
            description=description,
            parent_version=self.version,
        )



import logging

import os
import copy
from typing import Dict, Union
import requests
import time
import json
import threading
from datetime import datetime

from requests.models import Response

import previsionio.utils
from .logger import logger, event_logger
from . import config
from .utils import handle_error_response, parse_json, PrevisionException


class EventManager:
    def __init__(self, event_endpoint: str, auth_headers: Dict[str, str], client: 'Client'):
        self.event_endpoint = event_endpoint
        auth_headers = copy.deepcopy(auth_headers)
        self.headers = auth_headers
        self.client = client
        self.t = threading.Thread(target=self.update_events, daemon=True)
        self.t.start()

        self._events = (threading.Semaphore(1), {})

    @property
    def events(self):
        events_dict_copy = {}
        semd, event_dict = self._events

        with semd:
            for key, (semi, event_list) in event_dict.items():
                with semi:
                    events_dict_copy[key] = copy.deepcopy(event_list)

        return events_dict_copy

    def wait_for_event(self,
                       resource_id: str,
                       resource_type: str,
                       event_tuple: previsionio.utils.EventTuple,
                       specific_url: str):
        # ignore invalid resource ids
        if not isinstance(resource_id, str):
            return
        self.register_resource(resource_id)
        t0 = time.time()
        while time.time() < t0 + config.default_timeout:
            reconnect_start = time.time()
            semd, event_dict = self._events
            semd.acquire()
            if self.check_resource_event(resource_id, specific_url, event_tuple, semd):
                return
            semd.release()
            time.sleep(0.1)
            while time.time() < reconnect_start + 300:
                semd, event_dict = self._events
                semd.acquire()
                semi, event_list = event_dict[resource_id]
                remaining_events = []
                with semi:
                    for event in event_list:
                        name = event.get('event')
                        if name == event_tuple.name:
                            if self.check_resource_event(resource_id, specific_url, event_tuple, semd):
                                return
                        else:
                            remaining_events.append(event)

                event_dict[resource_id] = semi, remaining_events
                semd.release()
                time.sleep(0.1)
        else:
            raise TimeoutError('Failed to get status {} on {} {}'.format(event_tuple,
                                                                         resource_type,
                                                                         resource_id))

    def check_resource_event(self, resource_id: str, endpoint: str, event_tuple: previsionio.utils.EventTuple,
                             semd: threading.Semaphore):
        resp = self.client.request(endpoint=endpoint, method=requests.get, check_response=False)
        json_response = parse_json(resp)
        logging.debug("{} - endpoint='{}' -> '{}".format(datetime.now(), endpoint, json_response))
        if event_tuple.is_failure(json=json_response):
            semd.release()
            msg = 'Error on resource {}: {}\n{}'.format(resource_id,
                                                        json_response.get('errorMessage', ''),
                                                        json_response)
            raise PrevisionException(msg)
        if event_tuple.is_success(json=json_response):
            semd.release()
            return True
        return False

    def register_resource(self, resource_id: str):
        event_logger.debug('Registering resource with id {}'.format(resource_id))
        payload = {'event': 'REGISTER', '_id': resource_id}
        self.add_event(resource_id, payload)

    def update_events(self):
        sse_timeout = 300
        event_data = None
        event_name = None

        while True:
            sse = requests.get(self.event_endpoint,
                               stream=True,
                               headers=self.headers,
                               timeout=sse_timeout)

            try:
                for line in sse.iter_lines(chunk_size=None):
                    if not line:
                        # filter out keep-alive new lines
                        continue

                    event_logger.debug('url: {} -- data: {}'.format(self.event_endpoint, line))
                    line = line.decode()
                    if line.startswith(':'):
                        # SSE comments can start with ":" character
                        event_logger.debug('sse comment{}'.format(line))
                        continue
                    elif line.startswith('id '):
                        # skip id message information
                        continue
                    elif line.startswith('event: '):
                        # get event name
                        event_name = line[len('event: '):]
                    elif line.startswith('data: '):
                        if event_name is None:
                            # ignore event with unknown name
                            continue

                        # get event data
                        try:
                            event_data = json.loads(line[len('data: '):])
                        except Exception as e:
                            # skip malformed event data
                            event_logger.warning('failed to parse json: "{}" -- error: {}'.format(line, e.__repr__()))
                            continue

                        resource_id = event_data.get('_id', None)

                        if not isinstance(resource_id, str):
                            # ignore event with invalid resource id
                            continue

                        # add event only if monitored resource
                        semd, event_dict = self._events
                        if resource_id in event_dict:
                            payload = {'event': event_name, 'id': resource_id}
                            event_logger.debug('url: {} -- event: {} payload: {}'.format(self.event_endpoint,
                                                                                         event_name,
                                                                                         payload))
                            self.add_event(resource_id, payload)
            except requests.exceptions.ChunkedEncodingError:
                event_logger.warning('closing connection to endpoint: "{}"'.format(self.event_endpoint))
            except requests.exceptions.ConnectionError:
                logger.warning('{}: no messages in {} seconds. reconnecting'.format(self.event_endpoint, sse_timeout))
            except Exception as e:
                logger.error(e)
            finally:
                sse.close()

    def add_event(self, resource_id: str, payload: Union[None, Dict]):
        event_logger.debug('adding event for {}:\npayload = {}'.format(resource_id, payload))
        semd, event_dict = self._events

        if payload and isinstance(payload, dict):
            with semd:
                if resource_id in event_dict:
                    semi, event_list = event_dict[resource_id]
                else:
                    semi, event_list = (threading.Semaphore(1), [])

                with semi:
                    event_list.append(payload)

                event_dict[resource_id] = semi, event_list

                self._events = semd, event_dict


class Client(object):

    """Client class to interact with the Prevision.io platform and manage authentication."""

    def __init__(self):
        self.token = None
        self.prevision_url = None
        self.user_info = None
        self.headers = {
            'accept-charset': 'UTF-8',
            'cache-control': 'no-cache',
            'accept': 'application/json',
        }

        self.api_version = '/ext/v1'

        self.url = None

        self.event_manager = None

    def _check_token_url(self):

        if not self.token:
            raise PrevisionException('No token configured. Call client.init_client() to initialize')

        if not self.prevision_url:
            raise PrevisionException('No url configured. Call client.init_client() to initialize')

    def request(self, endpoint: str, method, files: Dict = None, data: Dict = None,
                format: Dict = None, allow_redirects: bool = True,
                content_type: str = None, check_response: bool = True,
                message_prefix: str = None, **requests_kwargs) -> Response:
        """
        Make a request on the desired endpoint with the specified method & data.

        Requires initialization.

        Args:
            endpoint: (str): api endpoint (e.g. /experiments, /prediction/file)
            method (requests.{get,post,delete}): requests method
            files (dict): files dict
            data (dict): for single predict
            content_type (str): force request content-type
            allow_redirects (bool): passed to requests method
            check_response (bool): wether to handle error or not
            message_prefix (str): prefix message in error logs

        Returns:
            request response

        Raises:
            Exception: Error if url/token not configured
        """
        self._check_token_url()

        headers = copy.deepcopy(self.headers)
        if content_type:
            headers['content-type'] = content_type

        if self.url is None:
            raise RuntimeError("client.url not properly initialized")

        if format and len(format):
            def format_bool(val):
                if isinstance(val, bool):
                    if val:
                        return "true"
                    return "false"
                return val
            endpoint += '?' + '&'.join(["{}={}".format(key, format_bool(val)) for key, val in format.items()])

        url = self.url + endpoint

        status_code = 502
        retries = config.request_retries
        n_tries = 0
        resp = None
        while (n_tries < retries) and (status_code in config.retry_codes):
            n_tries += 1

            try:
                resp = method(url,
                              headers=headers,
                              files=files,
                              allow_redirects=allow_redirects,
                              json=data,
                              **requests_kwargs)
                status_code = resp.status_code

            except Exception as e:
                raise PrevisionException(f'Error requesting: {url} with error {e.__repr__()}')

            if status_code in config.retry_codes:
                logger.warning(f'Failed to request {url} with status code {status_code}.'
                               f' Retrying {retries - n_tries} times')
                time.sleep(config.request_retry_time)

        assert isinstance(resp, Response)
        if check_response:
            handle_error_response(resp=resp,
                                  url=url,
                                  data=data,
                                  files=files,
                                  message_prefix=message_prefix,
                                  n_tries=n_tries)

        return resp

    def update_user_info(self):
        assert self.url is not None
        user_info_response = requests.get(self.url + '/profile',
                                          headers=self.headers)
        result = parse_json(user_info_response)

        if 'err_code' in result and result['err_code'] == 'E_UNK':
            raise ValueError('Wrong token ' + str(result))
        self.user_info = result

    def init_client(self, prevision_url: str, token: str):
        """
        Init the client (and check that the connection is valid).

        Args:
            prevision_url (str): URL of the Prevision.io platform. Should be of the form
                https://<instance_name>.prevision.io, or a custom IP address if working on-premise.

            token (str): Your Prevision.io master token. Can be retrieved on
                /dashboard/infos on the web interface or obtained programmatically through:

                .. code-block:: python

                    client.init_client(prevision_url, token)
        """
        self.prevision_url = prevision_url
        self.url = self.prevision_url + self.api_version
        self.token = token
        self.headers['Authorization'] = self.token

        # check for correct connection
        resp = self.request('/version', requests.get)
        handle_error_response(resp, '/version')

        logger.debug('subscribing to events manager')
        self.event_manager = EventManager(self.url + '/events', auth_headers=self.headers, client=self)


client = Client()


if os.getenv('PREVISION_URL') and os.getenv('PREVISION_MASTER_TOKEN'):
    logger.info('Initializing Prevision.io client using environment variables')
    logger.debug('PREVISION_URL:' + os.getenv('PREVISION_URL', ""))
    logger.debug('PREVISION_MASTER_TOKEN:' + os.getenv('PREVISION_MASTER_TOKEN', ""))
    client.init_client(os.getenv('PREVISION_URL', ""), os.getenv('PREVISION_MASTER_TOKEN', ""))



from typing import List
import time
import requests

from . import config
from .logger import logger
from .api_resource import ApiResource
from . import client
from .experiment_version import BaseExperimentVersion
from .utils import parse_json, PrevisionException, get_all_results
from .prediction import DeploymentPrediction
from .dataset import Dataset
from .model import Model


class ExperimentDeployment(ApiResource):
    """ ExperimentDeployment objects represent experiment deployment
    resource that will be explored by Prevision.io platform.
    """

    resource = 'model-deployments'

    def __init__(self, _id: str, name: str, experiment_id, current_version,
                 versions, deploy_state, access_type, project_id, training_type, models, url=None,
                 **kwargs):

        self.name = name
        self._id = _id
        # self.experiment_version_id = experiment_version_id
        self.experiment_id = experiment_id
        self.current_version = current_version
        self.versions = versions
        self._deploy_state = deploy_state
        self.access_type = access_type
        self.project_id = project_id
        self.training_type = training_type
        self.models = models
        self.url = url

        self._run_state = kwargs.pop("main_model_run_state", kwargs.pop("run_state", "error"))
        for k, v in kwargs.items():
            self.__setattr__(k, v)

    @classmethod
    def from_id(cls, _id: str):
        """Get a deployed experiment from the platform by its unique id.

        Args:
            _id (str): Unique id of the experiment version to retrieve

        Returns:
            :class:`.ExperimentDeployment`: Fetched deployed experiment

        Raises:
            PrevisionException: Any error while fetching data from the platform
                or parsing result
        """
        url = '/{}/{}'.format('deployments', _id)
        result = super()._from_id(_id=_id, specific_url=url)
        logger.debug(result)
        return cls(**result)

    @property
    def deploy_state(self):
        experiment_deployment = self.from_id(self._id)
        return experiment_deployment._deploy_state

    @property
    def run_state(self):
        experiment_deployment = self.from_id(self._id)
        return experiment_deployment._run_state

    @classmethod
    def list(cls, project_id: str, all: bool = True) -> List['ExperimentDeployment']:
        """ List all the available experiment in the current active [client] workspace.

        .. warning::

            Contrary to the parent ``list()`` function, this method
            returns actual :class:`.ExperimentDeployment` objects rather
            than plain dictionaries with the corresponding data.

        Args:
            project_id (str): project id
            all (boolean, optional): Whether to force the SDK to load all items of
                the given type (by calling the paginated API several times). Else,
                the query will only return the first page of result.

        Returns:
            list(:class:`.ExperimentDeployment`): Fetched dataset objects
        """
        resources = super()._list(all=all, project_id=project_id)
        return [ExperimentDeployment(**experiment_deployment) for experiment_deployment in resources]

    @classmethod
    def _new(
        cls,
        project_id: str,
        name: str,
        main_model: Model,
        challenger_model: Model = None,
        access_type: str = 'public',
    ):
        """ Create a new experiment deployment object on the platform.

        Args:
            project_id (str): project id
            name (str): experiment deployment name
            main_model (:class:`.Model`): main model
            challenger_model (:class:`.Model`, optional): challenger model (main and challenger
                models should come from the same experiment)
            access_type (str, optional): public/ fine_grained/ private

        Returns:
            :class:`.ExperimentDeployment`: The registered experiment deployment object in the current project

        Raises:
            PrevisionException: Any error while creating experiment deployment to the platform
                or parsing the result
            Exception: For any other unknown error
        """

        if access_type not in ['public', 'fine_grained', 'private']:
            raise PrevisionException('access type must be public, fine_grained or private')
        main_model_experiment_version_id = main_model.experiment_version_id
        main_experiment = BaseExperimentVersion._from_id(main_model_experiment_version_id)
        main_experiment_id = main_experiment['experiment_id']
        data = {
            'name': name,
            'experiment_id': main_experiment_id,
            'main_model_experiment_version_id': main_model_experiment_version_id,
            'main_model_id': main_model._id,
            'access_type': access_type
        }

        if challenger_model:
            challenger_model_experiment_version_id = challenger_model.experiment_version_id
            challenger_experiment = BaseExperimentVersion._from_id(main_model_experiment_version_id)
            challenger_experiment_id = challenger_experiment['experiment_id']
            if main_experiment_id != challenger_experiment_id:
                raise PrevisionException('main and challenger models must be from the same experiment')
            data['challenger_model_experiment_version_id'] = challenger_model_experiment_version_id
            data['challenger_model_id'] = challenger_model._id

        url = '/projects/{}/model-deployments'.format(project_id)
        resp = client.request(url,
                              data=data,
                              method=requests.post,
                              message_prefix='ExperimentDeployment creation')
        json_resp = parse_json(resp)
        experiment_deployment = cls.from_id(json_resp['_id'])
        return experiment_deployment

    def new_version(self, name: str, main_model, challenger_model=None):
        """ Create a new experiment deployment version.

        Args:
            name (str): experiment deployment name
            main_model: main model
            challenger_model (optional): challenger model. main and challenger models should be in the same experiment

        Returns:
            :class:`.ExperimentDeployment`: The registered experiment deployment object in the current project

        Raises:
            PrevisionException: Any error while creating experiment deployment to the platform
                or parsing the result
            Exception: For any other unknown error
        """
        main_model_experiment_version_id = main_model.experiment_version_id
        main_experiment = BaseExperimentVersion._from_id(main_model_experiment_version_id)
        main_experiment_id = main_experiment['experiment_id']
        data = {
            'name': name,
            'main_model_experiment_version_id': main_model_experiment_version_id,
            'main_model_id': main_model._id,
        }
        if challenger_model:
            challenger_model_experiment_version_id = challenger_model.experiment_version_id
            challenger_experiment = BaseExperimentVersion._from_id(main_model_experiment_version_id)
            challenger_experiment_id = challenger_experiment['experiment_id']
            if main_experiment_id != challenger_experiment_id:
                raise PrevisionException('main and challenger models must be from the same experiment')
            data['challenger_model_experiment_version_id'] = challenger_model_experiment_version_id
            data['challenger_model_id'] = challenger_model._id

        url = '/deployments/{}/versions'.format(self._id)
        resp = client.request(url,
                              data=data,
                              method=requests.post,
                              message_prefix='ExperimentDeployment creation new version')
        json_resp = parse_json(resp)
        experiment_deployment = self.from_id(json_resp['_id'])
        return experiment_deployment

    def delete(self):
        """Delete an experiment deployment from the actual [client] workspace.

        Raises:
            PrevisionException: If the experiment deployment does not exist
            requests.exceptions.ConnectionError: Error processing the request
        """
        super().delete()

    def wait_until(self, condition, timeout: float = config.default_timeout):
        """ Wait until condition is fulfilled, then break.

        Args:
            condition (func: (:class:`.BaseExperimentVersion`) -> bool.): Function to use to check the
                break condition
            raise_on_error (bool, optional): If true then the function will stop on error,
                otherwise it will continue waiting (default: ``True``)
            timeout (float, optional): Maximal amount of time to wait before forcing exit

        Example::

            experiment.wait_until(lambda experimentv: len(experimentv.models) > 3)

        Raises:
            PrevisionException: If the resource could not be fetched or there was a timeout.
        """
        t0 = time.time()
        while True:
            if timeout is not None and time.time() - t0 > timeout:
                raise PrevisionException('timeout while waiting on {}'.format(condition))
            try:
                if condition(self):
                    break
                elif self.deploy_state == 'failed' or self.run_state == 'failed':
                    raise PrevisionException('Resource failed while waiting')
            except PrevisionException as e:
                logger.warning(e.__repr__())
                raise

            time.sleep(config.scheduler_refresh_rate)

    def create_api_key(self):
        """Create an api key of the experiment deployment from the actual [client] workspace.

        Raises:
            PrevisionException: If the dataset does not exist
            requests.exceptions.ConnectionError: Error processing the request
        """
        print('/deployments/{}/api-keys'.format(self.id))
        resp = client.request(endpoint='/deployments/{}/api-keys'.format(self.id),
                              method=requests.post,
                              message_prefix='ExperimentDeployment create api key')
        resp = parse_json(resp)
        return resp

    def get_api_keys(self):
        """Fetch the api keys client id and cient secret of the
        experiment deployment from the actual [client] workspace.

        Raises:
            PrevisionException: If the dataset does not exist
            requests.exceptions.ConnectionError: Error processing the request
        """
        resp = client.request(endpoint='/deployments/{}/api-keys'.format(self.id),
                              method=requests.get,
                              message_prefix='ExperimentDeployment get api key _id')
        resp = parse_json(resp)
        api_keys_ids = [item['_id'] for item in resp['items']]
        res = []
        for api_keys_id in api_keys_ids:
            resp = client.request(endpoint='/api-keys/{}/secret'.format(api_keys_id),
                                  method=requests.get,
                                  message_prefix='ExperimentDeployment get api key client_id and secret')
            resp = parse_json(resp)
            res.append({'client_id': resp['service_account_client_id'],
                        'client_secret': resp['client_secret']})
        return res

    def predict_from_dataset(self, dataset: Dataset) -> DeploymentPrediction:
        """ Make a prediction for a dataset stored in the current active [client]
        workspace (using the current SDK dataset object).

        Args:
            dataset (:class:`.Dataset`): Dataset resource to make a prediction for

        Returns:
            :class:`.DeploymentPrediction`: The registered prediction object in the current workspace
        """
        if self.training_type not in ['regression', 'classification', 'multiclassification']:
            PrevisionException('Prediction not supported yet for training type {}', self.training_type)
        data = {
            'dataset_id': dataset._id,
        }

        predict_start = client.request('/deployments/{}/deployment-predictions'.format(self._id),
                                       method=requests.post,
                                       data=data,
                                       message_prefix='Bulk predict')
        predict_start_parsed = parse_json(predict_start)
        return DeploymentPrediction(**predict_start_parsed)

    def list_predictions(self) -> List[DeploymentPrediction]:
        """ List all the available predictions in the current active [client] workspace.

        Returns:
            list(:class:`.DeploymentPrediction`): Fetched deployed predictions objects
        """
        end_point = '/deployments/{}/deployment-predictions'.format(self._id)
        predictions = get_all_results(client, end_point, method=requests.get)
        return [DeploymentPrediction(**prediction) for prediction in predictions]



import logging
from previsionio.logger import logger, event_logger


__version__ = '11.2.0'


def verbose(v, debug: bool = False, event_log: bool = False):
    """ Set the SDK level of verbosity.

    Args:
        v (bool): Whether to activate info logging
        debug (bool, optional): Whether to activate detailed
            debug logging (default: ``False``)
        event_log (bool, optional): Whether to activate detailed
            event managers debug logging (default: ``False``)
    """
    if event_log:
        event_logger.setLevel(logging.DEBUG)

    if v:
        if debug:
            logger.setLevel(logging.DEBUG)
        else:
            logger.setLevel(logging.INFO)
    else:
        logger.setLevel(logging.WARNING)


verbose(False)


class Config:
    def __init__(self):
        self.auto_update: bool = True
        self.zip_files: bool = True
        self.success_codes: list = list(range(200, 211)) + [226]
        self.retry_codes: list = [500, 502, 503, 504]
        self.request_retries: int = 6
        self.request_retry_time: int = 10
        self.scheduler_refresh_rate: int = 10
        self.default_timeout: float = 3600.


config = Config()

from previsionio.prevision_client import client
from previsionio.experiment_config import (
    AdvancedModel,
    NormalModel,
    SimpleModel,
    TypeProblem,
    Feature,
    TrainingConfig,
    Profile,
    base_config,
    quick_config,
    ultra_config,
    ColumnConfig,
    YesOrNo,
    YesOrNoOrAuto,
)

from previsionio.connector import Connector
from previsionio.datasource import DataSource
from previsionio.exporter import Exporter, ExporterWriteMode
from previsionio.export import Export
from previsionio.project import Project
from previsionio.experiment import Experiment
from previsionio.supervised import Supervised
from previsionio.timeseries import (
    TimeSeries,
    TimeWindow,
    TimeWindowException,
)
from previsionio.model import (
    Model,
    ClassificationModel,
    RegressionModel,
    MultiClassificationModel,
    TextSimilarityModel,
)
from previsionio.text_similarity import (
    TextSimilarity,
    DescriptionsColumnConfig,
    QueriesColumnConfig,
    ListModelsParameters,
    ModelsParameters,
    TextSimilarityModels,
    ModelEmbedding,
    Preprocessing,
)
from previsionio.dataset import Dataset, DatasetImages
from previsionio.deployed_model import DeployedModel
from previsionio.experiment_deployment import ExperimentDeployment
from previsionio.prediction import ValidationPrediction, DeploymentPrediction

from previsionio.pipeline import PipelineScheduledRun

__all__ = [
    'client',
    'AdvancedModel',
    'NormalModel',
    'SimpleModel',
    'TypeProblem',
    'Feature',
    'TrainingConfig',
    'YesOrNo',
    'YesOrNoOrAuto',
    'Profile',
    'base_config',
    'quick_config',
    'ultra_config',
    'ColumnConfig',
    'Connector',
    'DataSource',
    'Exporter',
    'ExporterWriteMode',
    'Export',
    'Supervised',
    'TimeSeries',
    'TimeWindow',
    'TimeWindowException',
    'Dataset',
    'DatasetImages',
    'Model',
    'RegressionModel',
    'ClassificationModel',
    'MultiClassificationModel',
    'TextSimilarityModel',
    'DeployedModel',
    'TextSimilarity',
    'DescriptionsColumnConfig',
    'QueriesColumnConfig',
    'ListModelsParameters',
    'ModelsParameters',
    'TextSimilarityModels',
    'ModelEmbedding',
    'Preprocessing',
    'Project',
    'Experiment',
    'ExperimentDeployment',
    'ValidationPrediction',
    'DeploymentPrediction',
    'PipelineScheduledRun'
]

from . import _version
__version__ = _version.get_versions()['version']



# -*- coding: utf-8 -*-
from previsionio.text_similarity import TextSimilarity
from previsionio.supervised import Supervised
from previsionio.timeseries import TimeSeries
from previsionio.external_experiment_version import ExternalExperimentVersion
from previsionio.experiment_config import DataType, TypeProblem
from typing import Dict, List, Union
import requests

from .prevision_client import client
from .utils import parse_json
from .api_resource import ApiResource


def get_experiment_version_class(
    training_type: TypeProblem,
    data_type: DataType,
    provider: str,
) -> Union[TextSimilarity, Supervised, TimeSeries, ExternalExperimentVersion]:
    """ Get the type of ExperimentVersion class used by this Experiment

    Returns:
        (:class:`.previsionio.text_similarity.TextSimilarity` | :class:`.Supervised` |
        :class:`.TimeSeries` | :class:`.ExternalExperimentVersion`):
        Type of ExperimentVersion
    """
    if provider == 'external':
        experiment_version_class = ExternalExperimentVersion
    else:
        if training_type == TypeProblem.TextSimilarity:
            experiment_version_class = TextSimilarity
        else:
            if data_type == DataType.TimeSeries:
                experiment_version_class = TimeSeries
            elif data_type in [DataType.Tabular, DataType.Images]:
                experiment_version_class = Supervised
            else:
                raise ValueError('There is no experiment_version_class with the following values: '
                                 f'training_type: {training_type}'
                                 f'data_type: {data_type}'
                                 f'provider: {provider}')
    return experiment_version_class


class Experiment(ApiResource):
    """ An Experiment

    Args:
        _id (str): Unique id of the experiment
        name (str): Name of the experiment

    """

    resource = 'experiments'

    def __init__(self,
                 _id: str,
                 project_id: str,
                 provider: str,
                 name: str,
                 training_type: TypeProblem,
                 data_type: DataType) -> 'Experiment':
        super().__init__(_id=_id)
        self.project_id = project_id
        self.provider = provider
        self.name = name
        self.training_type: TypeProblem = TypeProblem(training_type)
        self.data_type: DataType = DataType(data_type)

    # TODO: build a class enum for possible providers
    @classmethod
    def new(cls,
            project_id: str,
            provider: str,
            name: str,
            data_type: DataType,
            training_type: TypeProblem) -> 'Experiment':
        url = f'/projects/{project_id}/experiments'
        data = {
            'provider': provider,
            'name': name,
            'data_type': data_type.value,
            'training_type': training_type.value,
        }
        response = client.request(url,
                                  method=requests.post,
                                  data=data,
                                  message_prefix='Experiment creation')
        experiment_info = parse_json(response)
        experiment = cls.from_dict(experiment_info)
        return experiment

    @classmethod
    def from_dict(cls, experiment_info: Dict) -> 'Experiment':
        experiment = cls(
            experiment_info['_id'],
            experiment_info['project_id'],
            experiment_info['provider'],
            experiment_info['name'],
            experiment_info['training_type'],
            experiment_info['data_type'],
        )
        return experiment

    @classmethod
    def from_id(cls, _id: str) -> 'Experiment':
        """Get an experiment from the platform by its unique id.

        Args:
            _id (str): Unique id of the experiment version to retrieve

        Returns:
            :class:`.Experiment`: Fetched experiment

        Raises:
            PrevisionException: Any error while fetching data from the platform
                or parsing result
        """
        experiment_info = super()._from_id(_id)
        experiment = cls.from_dict(experiment_info)
        return experiment

    @classmethod
    def list(cls, project_id: str, all: bool = True) -> List['Experiment']:
        """ List all the available experiment in the current active [client] workspace.

        .. warning::

            Contrary to the parent ``list()`` function, this method
            returns actual :class:`.Experiment` objects rather than
            plain dictionaries with the corresponding data.

        Args:
            project_id (str): project id
            all (boolean, optional): Whether to force the SDK to load all items of
                the given type (by calling the paginated API several times). Else,
                the query will only return the first page of result.

        Returns:
            list(:class:`.Experiment`): Fetched dataset objects
        """
        experiment_infos = super()._list(all=all, project_id=project_id)
        return [cls.from_dict(experiment_info) for experiment_info in experiment_infos]

    @property
    def experiment_version_class(self) -> Union[TextSimilarity, Supervised, TimeSeries, ExternalExperimentVersion]:
        """ Get the type of ExperimentVersion class used by this Experiment

        Returns:
            (:class:`previsionio.text_similarity.TextSimilarity` | :class:`.Supervised` |
            :class:`.TimeSeries` | :class:`.ExternalExperimentVersion`):
            Type of ExperimentVersion
        """
        return get_experiment_version_class(self.training_type, self.data_type, self.provider)

    @property
    def latest_version(self) -> Union[TextSimilarity, Supervised, TimeSeries]:
        """Get the latest version of this experiment version.

        Returns:
            (:class:`previsionio.text_similarity.TextSimilarity` | :class:`.Supervised` | :class:`.TimeSeries`):
            latest ExperimentVersion in this Experiment
        """
        end_point = '/{}/{}/versions'.format(self.resource, self._id)
        format = {
            "rowsPerPage": 1,
            "sortBy": "created_at",
            "descending": True
        }

        response = client.request(endpoint=end_point,
                                  format=format,
                                  method=requests.get,
                                  message_prefix='Latest experiment version')
        res = parse_json(response)
        assert len(res['items']) == 1
        return self.experiment_version_class(**res['items'][0])

    @property
    def versions(self) -> List[Union[TextSimilarity, Supervised, TimeSeries]]:
        """Get the list of all versions for the current experiment version.

        Returns:
            list(:class:`previsionio.text_similarity.TextSimilarity` | :class:`.Supervised` | :class:`.TimeSeries`):
            List of the experiment versions (as JSON metadata)
        """
        end_point = '/{}/{}/versions'.format(self.resource, self._id)
        response = client.request(endpoint=end_point,
                                  method=requests.get,
                                  message_prefix='Experiment versions listing')
        res = parse_json(response)
        return [self.experiment_version_class(**val) for val in res['items']]

    def delete(self):
        """Delete an experiment from the actual [client] workspace.

        Raises:
            PrevisionException: If the experiment does not exist
            requests.exceptions.ConnectionError: Error processing the request
        """
        super().delete()



import requests
from typing import Dict, List
from .api_resource import ApiResource
from . import client
from .logger import logger
from .utils import parse_json


class PipelineScheduledRun(ApiResource):
    resource = 'pipeline-scheduled-runs'

    def __init__(
        self,
        _id: str,
        name: str,
        project_id: str,
        pipeline_template_id: Dict,
        nodes_properties: List,
        exec_type: str,
        enabled: bool,
        description: str = None,
        draft: bool = True,
        created_at: str = None,
        **kwargs
    ):
        """ A Pipeline Scheduled Run """
        super().__init__(_id=_id)
        self._id = _id
        self.name = name
        self.project_id = project_id
        self.pipeline_template_id = pipeline_template_id
        self.nodes_properties = nodes_properties
        self.exec_type = exec_type
        self.enabled = enabled
        self.description = description
        self.draft = draft
        self.created_at = created_at

    @classmethod
    def from_id(cls, _id: str) -> 'PipelineScheduledRun':
        """ Get a pipeline scheduled run from the platform by its unique id.

        Args:
            _id (str): Unique id of the pipeline scheduled run to retrieve

        Returns:
            :class:`.PipelineScheduledRun`: Fetched experiment

        Raises:
            PrevisionException: Any error while fetching data from the platform
                or parsing result
        """
        return cls(**super()._from_id(_id=_id))

    @classmethod
    def list(cls, project_id: str, all: bool = True):
        """ List all the available pipeline scheduled runs in the current active [client] workspace.

        .. warning::

            Contrary to the parent ``list()`` function, this method
            returns actual :class:`.Dataset` objects rather than
            plain dictionaries with the corresponding data.

        Args:
            project_id (str): Unique reference of the project id on the platform
            all (boolean, optional): Whether to force the SDK to load all items of
                the given type (by calling the paginated API several times). Else,
                the query will only return the first page of result.

        Returns:
            list(:class:`.PipelineScheduledRun`): Fetched dataset objects
        """
        resources = super()._list(all=all, project_id=project_id)
        return [cls(**conn_data) for conn_data in resources]

    def trigger(self):
        """ Trigger an execution of a pipeline scheduled run. """
        url = '/{}/{}/trigger'.format(self.resource, self._id)
        client.request(url, method=requests.post, message_prefix='PipelineScheduledRun trigger')
        logger.info('pipeline scheduled run with id {} triggered'.format(self._id))

    def get_executions(self, limit: int = 15):
        """Get executions of a pipeline scheduled run.

        Args:
            limit (int): Number of executions to retrieve.

        Returns:
            list(dict)
        """
        url = '/{}/{}?limit={}'.format(self.resource, self._id, limit)
        run_resp = client.request(url, method=requests.get, message_prefix='PipelineScheduledRun get element')
        run_json = parse_json(run_resp)
        return run_json['executions']



from setuptools import setup, find_packages

import versioneer


with open('requirements.txt', mode='r', encoding='utf-8') as f:
    install_requires = f.read().splitlines()

with open("README.md") as readme:
    setup(name='previsionio',
          version=versioneer.get_version(),
          cmdclass=versioneer.get_cmdclass(),
          description='python wrapper for prevision api',
          long_description=readme.read(),
          url='https://github.com/previsionio/prevision-python',
          author='prevision.io',
          author_email='team-core@prevision.io',
          packages=find_packages(exclude=['ci*', 'utest*']),
          license="MIT License",
          install_requires=install_requires,
          zip_safe=False,
          python_requires=">= 3.6",
          keywords='ml, ai, prevision, sdk',
          classifier=[
              'Intended Audience :: Developers',
              'Natural Language :: English',
              'License :: OSI Approved :: MIT License',
              'Programming Language :: Python :: 3',
              'Programming Language :: Python :: 3.6',
              'Programming Language :: Python :: 3.7',
              'Programming Language :: Python :: 3.8',
              'Programming Language :: Python :: 3.9',
          ]
          )



import requests
from enum import Enum
from typing import Union

from . import client
from .utils import parse_json, PrevisionException
from .api_resource import ApiResource, UniqueResourceMixin
from .connector import Connector, GCloud
from .export import Export
from .dataset import Dataset
from .prediction import DeploymentPrediction, ValidationPrediction


class ExporterWriteMode(Enum):
    """ Write mode for exporters.
    """
    safe = 'safe'
    """Fail if file already exists."""
    replace = 'replace'
    """Replace existing file/table."""
    append = 'append'
    """Append to existing table."""
    timestamp = 'timestamp'
    """Append timestamp to the output filename."""


class Exporter(ApiResource, UniqueResourceMixin):

    """ An exporter to access a distant data pool and upload data easily. This
    resource is linked to a :class:`.Connector` resource that represents the connection to
    the distant data source.

    Args:
        _id (str): Unique id of the exporter
        connector_id (str): Reference to the associated connector (the resource
            to go through to get a data snapshot)
        name (str): Name of the exporter
        description (str, optional): Description of the exporter
        path (str, optional): Path to the file to write on via the exporter
        bucket (str, optional): Bucket of the file to write on via the exporter
        database (str, optional): Name of the database to write on via the exporter
        table (str, optional): Name of the table to write on via the exporter
        g_cloud (:class:`.GCloud`, optional): Type of google cloud service
        write_mode (:class:`ExporterWriteMode`, optional): Write mode
    """

    resource = 'exporters'

    def __init__(self, _id, connector_id: str, name: str, description: str = None, path: str = None,
                 bucket: str = None, database: str = None, table: str = None, g_cloud: GCloud = None,
                 write_mode: ExporterWriteMode = ExporterWriteMode.safe, **kwargs):
        """ Instantiate a new :class:`.Exporter` object to manipulate an exporter resource
        on the platform. """
        super().__init__(_id=_id)

        self.connector_id = connector_id

        self.name = name
        self.description = description
        self.path = path
        self.bucket = bucket
        self.database = database
        self.table = table
        self.g_cloud = g_cloud
        self.write_mode = write_mode

        self.other_params = kwargs

    @classmethod
    def list(cls, project_id: str, all: bool = False):
        """ List all the available exporters in the current active [client] workspace.

        Args:
            project_id (str): Unique reference of the project id on the platform
            all (boolean, optional): Whether to force the SDK to load all items of
                the given type (by calling the paginated API several times). Else,
                the query will only return the first page of result.

        Returns:
            list(:class:`.Exporter`): The fetched exporter objects
        """
        resources = super()._list(all=all, project_id=project_id)
        return [cls(**source_data) for source_data in resources]

    @classmethod
    def from_id(cls, _id: str):
        """Get an exporter from the instance by its unique id.

        Args:
            _id (str): Unique id of the exporter to retrieve

        Returns:
            :class:`.Exporter`: The fetched exporter

        Raises:
            PrevisionException: Any error while fetching data from the platform
                or parsing the result
        """
        resp_json = super()._from_id(_id=_id)
        return cls(**resp_json)

    @classmethod
    def _new(cls, project_id: str, connector: Connector, name: str, description: str = None, path: str = None,
             bucket: str = None, database: str = None, table: str = None, g_cloud: GCloud = None,
             write_mode: ExporterWriteMode = ExporterWriteMode.safe):
        """ Create a new exporter object on the platform.

        Args:
            project_id (str): Unique project id on which to create the exporter
            connector (:class:`.Connector`): Reference to the associated connector (the resource
                to go through to get a data snapshot)
            name (str): Name of the exporter
            description (str, optional): Description of the exporter
            path (str, optional): Path to the file to write on via the exporter
            bucket (str, optional): Bucket of the file to write on via the exporter
            database (str, optional): Name of the database to write on via the exporter
            table (str, optional): Name of the table to write on via the exporter
            g_cloud (:class:`.GCloud`, optional): Type of google cloud service
            write_mode (:class:`ExporterWriteMode`, optional): Write mode

        Returns:
            :class:`.Exporter`: The registered exporter object in the current workspace

        Raises:
            PrevisionException: Any error while uploading data to the platform
                or parsing the result
            Exception: For any other unknown error
        """

        data = {
            'connector_id': connector._id,
            'name': name,
            'description': description,
            'filepath': path,
            'bucket': bucket,
            'database': database,
            'g_cloud': g_cloud,
            'table': table,
        }

        write_mode = write_mode.value
        if database is not None:
            if write_mode not in ['replace', 'append']:
                raise PrevisionException(f'Write mode "{write_mode}" is not compatible with database connector')
            data['database_write_mode'] = write_mode
        else:
            if write_mode not in ['timestamp', 'safe', 'replace']:
                raise PrevisionException(f'Write mode "{write_mode}" is not compatible with file connectors')
            data['file_write_mode'] = write_mode

        url = '/projects/{}/{}'.format(project_id, cls.resource)
        resp = client.request(url,
                              data=data,
                              method=requests.post,
                              message_prefix='Exporter creation')
        json = parse_json(resp)

        return cls(json['_id'], connector._id, name, description=description, path=path,
                   bucket=bucket, database=database, table=table, write_mode=write_mode,
                   g_cloud=g_cloud)

    def delete(self):
        """Delete an exporter from the actual [client] workspace.

        Raises:
            PrevisionException: If the exporter does not exist
            requests.exceptions.ConnectionError: Error processing the request
        """
        super().delete()

    def export_file(self, file_path: str, encoding: str = None, separator: str = None, decimal: str = None,
                    thousands: str = None, wait_for_export: bool = False, **kwargs):
        """ Upload a CSV file using the exporter.

        Args:
            file_path (str): Path of the file to upload
            encoding (str, optional): Encoding of the file to upload
            separator (str, optional): Separator of the file to upload
            decimal (str, optional): Decimal of the file to upload
            thousands (str, optional): Thousands of the file to upload
            wait_for_export (bool, optional): Wether to wait until the export is complete or not

        Returns:
            :class:`.Export`: The registered export object
        """
        return Export.export_file(self._id, file_path=file_path, encoding=encoding, separator=separator,
                                  decimal=decimal, thousands=thousands, wait_for_export=wait_for_export,
                                  **kwargs)

    def export_dataset(self, dataset: Dataset, wait_for_export: bool = False):
        """ Upload a :class:`.Dataset` from the current active project using the exporter.

        Args:
            dataset (:class:`.Dataset`): dataset to upload
            wait_for_export (bool, optional): Wether to wait until the export is complete or not

        Returns:
            :class:`.Export`: The registered export object
        """
        return Export.export_dataset(exporter_id=self._id, dataset=dataset, wait_for_export=wait_for_export)

    def export_prediction(self, prediction: Union[DeploymentPrediction, ValidationPrediction],
                          wait_for_export: bool = False):
        """ Upload a :class:`.DeploymentPrediction` or a :class:`.ValidationPrediction`
        from the current active project using the exporter.

        Args:
            dataset (:class:`.DeploymentPrediction`|:class:`.ValidationPrediction`): prediction to upload
            wait_for_export (bool, optional): Wether to wait until the export is complete or not

        Returns:
            :class:`.Export`: The registered export object
        """
        return Export.export_prediction(self._id, prediction=prediction, wait_for_export=wait_for_export)

    def list_exports(self):
        """ List all the available exports given the exporter id.

        Returns:
            list(:class:`.Export`): The fetched export objects
        """
        return Export.list(self._id)



# -*- coding: utf-8 -*-
from __future__ import print_function

from typing import Dict
from . import TrainingConfig
from .experiment_config import DataType, ExperimentConfig, ColumnConfig, TypeProblem
from .experiment_version import ClassicExperimentVersion
from .metrics import Regression
from .model import RegressionModel
from .dataset import Dataset
from .utils import to_json


class TimeWindowException(Exception):
    pass


class TimeWindow(ExperimentConfig):
    """
    A time window object for representing either feature derivation window periods or
    forecast window periods.

    Args:
        derivation_start (int): Start of the derivation window (must be < 0)
        derivation_end (int): End of the derivation window (must be < 0)
        forecast_start (int): Start of the forecast window (must be > 0)
        forecast_end (int): End of the forecast window (must be > 0)
    """
    config = {
        'derivation_start': 'start_dw',
        'derivation_end': 'end_dw',
        'forecast_start': 'start_fw',
        'forecast_end': 'end_fw',
    }

    def __init__(self, derivation_start: int, derivation_end: int, forecast_start: int, forecast_end: int):
        """Instantiate a new :class:`.TimeWindow` util object."""

        if not derivation_start < derivation_end or not forecast_start < forecast_end:
            raise TimeWindowException('start must be smaller than end')

        if derivation_start > 0 or derivation_end > 0:
            raise TimeWindowException('derivation window bounds must be negative')

        if forecast_start < 0 or forecast_end < 0:
            raise TimeWindowException('forecast window bounds must be positive')

        if not(derivation_end <= ((2 * forecast_start) - forecast_end)):
            raise TimeWindowException('derivation_end must be smaller than (2 * forecast_start) - forecast_end')
        # Not valid anymore :
        # if not abs(derivation_end) >= (forecast_end - forecast_start):
        #     raise TimeWindowException('end of derivation window must be smaller than end of forecast window')

        self.derivation_start = derivation_start
        self.forecast_start = forecast_start
        self.derivation_end = derivation_end
        self.forecast_end = forecast_end


class TimeSeries(ClassicExperimentVersion):
    """ A supervised experiment version, for timeseries data """

    training_type = TypeProblem.Regression
    metric_type = Regression
    default_metric = Regression.RMSE
    data_type = DataType.TimeSeries
    model_class = RegressionModel

    def __init__(self, **experiment_version_info):
        super().__init__(**experiment_version_info)

    def _update_from_dict(self, **experiment_version_info):
        super()._update_from_dict(**experiment_version_info)
        timeseries_values = experiment_version_info['experiment_version_params']['timeseries_values']
        self.time_window = TimeWindow.from_dict(timeseries_values)

    @classmethod
    def from_id(cls, _id: str) -> 'TimeSeries':
        return cls(**super()._from_id(_id))

    @staticmethod
    def _build_experiment_version_creation_data(description, dataset, column_config, time_window, metric,
                                                holdout_dataset, training_config,
                                                parent_version=None) -> Dict:
        data = super(TimeSeries, TimeSeries)._build_experiment_version_creation_data(
            description,
            parent_version=parent_version,
        )

        data['dataset_id'] = dataset.id
        data.update(to_json(column_config))
        data.update(to_json(time_window))
        data['metric'] = metric if isinstance(metric, str) else metric.value
        data['holdout_dataset_id'] = holdout_dataset.id if holdout_dataset is not None else None
        data.update(to_json(training_config))

        return data

    @classmethod
    def _fit(
        cls,
        experiment_id: str,
        dataset: Dataset,
        column_config: ColumnConfig,
        time_window: TimeWindow,
        metric: Regression = None,
        holdout_dataset: Dataset = None,
        training_config: TrainingConfig = TrainingConfig(),
        description: str = None,
        parent_version: str = None,
    ) -> 'TimeSeries':
        """ Start a timeseries experiment version training with a specific configuration (on the platform).

        Args:
            experiment_id (str): The id of the experiment from which this version is created
            dataset (:class:`.Dataset`): Reference to the dataset
                object to use for as training dataset
            column_config (:class:`.ColumnConfig`): Column configuration for the experiment
                (see the documentation of the :class:`.ColumnConfig` resource for more details
                on each possible column types)
            time_window (:class:`.TimeWindow`): Time configuration
                (see the documentation of the :class:`.TimeWindow` resource for more details)
            metric (metrics.Regression): Specific metric to use for the experiment version
                (default ``None``)
            holdout_dataset (:class:`.Dataset`, optional): Reference to a dataset object to
                use as a holdout dataset (default: ``None``)
            training_config (:class:`.TrainingConfig`, optional): Specific training configuration
                (see the documentation of the :class:`.TrainingConfig` resource for more details
                on all the parameters)
            description (str, optional): The description of this experiment version (default: ``None``)
            parent_version (str, optional): The parent version of this experiment_version (default: ``None``)


        Returns:
            :class:`.TimeSeries`: Newly created supervised experiment version object
        """

        return super()._fit(
            experiment_id,
            description=description,
            parent_version=parent_version,
            dataset=dataset,
            column_config=column_config,
            time_window=time_window,
            metric=metric,
            holdout_dataset=holdout_dataset,
            training_config=training_config,
        )

    def new_version(
        self,
        dataset: Dataset = None,
        column_config: ColumnConfig = None,
        time_window: TimeWindow = None,
        metric: Regression = None,
        holdout_dataset: Dataset = None,
        training_config: TrainingConfig = None,
        description: str = None,
    ) -> 'TimeSeries':
        """
        Start a new timeseries experiment version training from this version (on the platform).
        The training parameters are copied from the current version and then overridden for those provided.

        Args:
            dataset (:class:`.Dataset`, optional): Reference to the dataset
                object to use for as training dataset
            column_config (:class:`.ColumnConfig`, optional): Column configuration for the experiment
                (see the documentation of the :class:`.ColumnConfig` resource for more details
                on each possible column types)
            metric (metrics.Regression, optional): Specific metric to use for the experiment version
            holdout_dataset (:class:`.Dataset`, optional): Reference to a dataset object to
                use as a holdout dataset
            training_config (:class:`.TrainingConfig`, optional): Specific training configuration
                (see the documentation of the :class:`.TrainingConfig` resource for more details
                on all the parameters)
            description (str, optional): The description of this experiment version (default: ``None``)
        Returns:
            :class:`.TimeSeries`: Newly created timeSeries experiment version object (new version)
        """
        return TimeSeries._fit(
            self.experiment_id,
            dataset if dataset is not None else self.dataset,
            column_config if column_config is not None else self.column_config,
            time_window if time_window is not None else self.time_window,
            metric if metric is not None else self.metric,
            holdout_dataset=holdout_dataset if holdout_dataset is not None else self.holdout_dataset,
            training_config=training_config if training_config is not None else self.training_config,
            description=description,
            parent_version=self.version,
        )



# -*- coding: utf-8 -*-
from __future__ import print_function
import os
from typing import Dict, List, Tuple, Union
import requests
import pandas as pd

from . import metrics
from .experiment_config import TypeProblem
from .experiment_version import ClassicExperimentVersion
from .model import ExternalRegressionModel, ExternalClassificationModel, ExternalMultiClassificationModel
from .prevision_client import client
from .utils import parse_json
from .dataset import Dataset


# NOTE: We inherit from ClassicExperimentVersion because it contains a lot of methods we need here, but not all,
#       we could make a big refactor with intermediate Base classes
class ExternalExperimentVersion(ClassicExperimentVersion):

    def __get_model_class(self):
        if self.training_type is None:
            model_class = None
        elif self.training_type == TypeProblem.Regression:
            model_class = ExternalRegressionModel
        elif self.training_type == TypeProblem.Classification:
            model_class = ExternalClassificationModel
        elif self.training_type == TypeProblem.MultiClassification:
            model_class = ExternalMultiClassificationModel
        else:
            raise ValueError(f'Unknown training_type for ExternalExperimentVersion: {self.training_type}')
        return model_class

    def __init__(self, **experiment_version_info):
        super().__init__(**experiment_version_info)

    def _update_from_dict(self, **experiment_version_info):
        # we don't want to inherit from ClassicExperimentVersion._update_from_dict but from its mother...
        super(ClassicExperimentVersion, self)._update_from_dict(**experiment_version_info)

        experiment_version_params = experiment_version_info['experiment_version_params']

        self.holdout_dataset_id: str = experiment_version_info['holdout_dataset_id']
        self.target_column = experiment_version_params['target_column']

        # this is dict, maybe we should parse it in a tuple like in creation
        self.external_models = experiment_version_info.get('external_models')

        self.metric = experiment_version_info.get('metric')
        self.dataset_id: Union[str, None] = experiment_version_info.get('dataset_id')

        self.metric: str = experiment_version_params['metric']

        self.model_class = self.__get_model_class()

    @property
    def holdout_dataset(self) -> Dataset:
        """ Get the :class:`.Dataset` object corresponding to the holdout dataset of this experiment version.

        Returns:
            :class:`.Dataset`: Associated holdout dataset
        """
        return Dataset.from_id(self.holdout_dataset_id)

    @property
    def dataset(self) -> Union[Dataset, None]:
        """ Get the :class:`.Dataset` object corresponding to the training dataset of this experiment version.

        Returns:
            :class:`.Dataset`: Associated training dataset
        """
        return Dataset.from_id(self.dataset_id) if self.dataset_id is not None else None

    def _update_draft(self, external_models, **kwargs):
        self.__add_external_models(external_models)

    @staticmethod
    def _build_experiment_version_creation_data(description, holdout_dataset, target_column,
                                                metric, dataset, parent_version=None,
                                                **kwargs) -> Dict:
        data = super(ExternalExperimentVersion, ExternalExperimentVersion)._build_experiment_version_creation_data(
            description,
            parent_version=parent_version,
        )

        data['holdout_dataset_id'] = holdout_dataset.id
        data['target_column'] = target_column
        data['metric'] = metric if isinstance(metric, str) else metric.value
        data['dataset_id'] = dataset.id if dataset is not None else None

        return data

    @classmethod
    def _fit(
        cls,
        experiment_id: str,
        holdout_dataset: Dataset,
        target_column: str,
        external_models: List[Tuple],
        metric: metrics.Enum,
        dataset: Dataset = None,
        description: str = None,
        parent_version: str = None,
    ) -> 'ExternalExperimentVersion':
        """ Create an external experiment version with a specific configuration (on the platform).

        Args:
            experiment_id (str): The id of the experiment from which this version is created
            holdout_dataset (:class:`.Dataset`): Reference to the holdout dataset object to use for as holdout dataset
            target_column (str): The name of the target column for this experiment version
            external_models (list(tuple)): The external models to add in the experiment version to create.
                Each tuple contains 3 items describing an external model as follows:

                    1) The name you want to give to the model
                    2) The path to the model in onnx format
                    3) The path to a yaml file containing metadata about the model
            metric (metrics.Enum): Specific metric to use for the experiment version
            dataset (:class:`.Dataset`, optional): Reference to the dataset object that
                has been used to train the model (default: ``None``)
            description (str, optional): The description of this experiment version (default: ``None``)
            parent_version (str, optional): The parent version of this experiment_version (default: ``None``)


        Returns:
            :class:`.ExternalExperimentVersion`: Newly created external experiment version object
        """
        return super()._fit(
            experiment_id,
            description=description,
            parent_version=parent_version,
            holdout_dataset=holdout_dataset,
            target_column=target_column,
            external_models=external_models,
            metric=metric,
            dataset=dataset,
        )

    def new_version(
        self,
        external_models: List[Tuple],
        holdout_dataset: Dataset = None,
        target_column: str = None,
        metric: metrics.Enum = None,
        dataset: Dataset = None,
        description: str = None,
    ) -> 'ExternalExperimentVersion':
        """
        Create a new external experiment version from this version (on the platform).
        The external_models parameter is mandatory.
        The other parameters are copied from the current version and then overridden for those provided.

        Args:
            external_models (list(tuple)): The external models to add in the experiment version to create.
                Each tuple contains 3 items describing an external model as follows:

                    1) The name you want to give to the model
                    2) The path to the model in onnx format
                    3) The path to a yaml file containing metadata about the model
            holdout_dataset (:class:`.Dataset`, optional): Reference to the holdout dataset object
                to use for as holdout dataset
            target_column (str, optional): The name of the target column for this experiment version
            metric (metrics.Enum, optional): Specific metric to use for the experiment version
            dataset (:class:`.Dataset`, optional): Reference to the dataset object that
                has been used to train the model (default: ``None``)
            description (str, optional): The description of this experiment version (default: ``None``)
        Returns:
            :class:`.ExternalExperimentVersion`: Newly created external experiment object (new version)
        """
        return ExternalExperimentVersion._fit(
            self.experiment_id,
            holdout_dataset if holdout_dataset is not None else self.holdout_dataset,
            target_column if target_column is not None else self.target_column,
            external_models,
            metric if metric is not None else self.metric,
            dataset=dataset if dataset is not None else self.dataset,
            description=description,
            parent_version=self.version,
        )

    def __add_external_model(self, external_model: Tuple) -> None:
        external_model_upload_endpoint = f'/experiment-versions/{self._id}/external-models'
        external_model_upload_method = requests.post
        external_model_message_prefix = 'External model uploading'
        name_key = 'name'
        onnx_key = 'onnx_file'
        yaml_key = 'yaml_file'
        onnx_content_type = 'application/octet-stream'
        yaml_content_type = 'text/x-yaml'

        name, onnx_file, yaml_file = external_model
        onnx_filename = os.path.basename(onnx_file)
        yaml_filename = os.path.basename(yaml_file)
        with open(onnx_file, 'rb') as onnx_fd, open(yaml_file, 'rb') as yaml_fd:
            external_model_upload_files = [
                (name_key, (None, name)),
                (onnx_key, (onnx_filename, onnx_fd, onnx_content_type)),
                (yaml_key, (yaml_filename, yaml_fd, yaml_content_type)),
            ]
            external_model_upload_response = client.request(external_model_upload_endpoint,
                                                            method=external_model_upload_method,
                                                            files=external_model_upload_files,
                                                            message_prefix=external_model_message_prefix,
                                                            )
        experiment_version_info = parse_json(external_model_upload_response)
        self._update_from_dict(**experiment_version_info)

    def __add_external_models(self, external_models: List[Tuple]) -> None:
        for external_model in external_models:
            self.__add_external_model(external_model)

    @property
    def train_dataset(self):
        raise NotImplementedError

    @property
    def best_model(self):
        """ Get the model with the best predictive performance over all models, where the best performance
        corresponds to a minimal loss.

        Returns:
            (:class:`.Model`, None): Model with the best performance in the experiment, or
            ``None`` if no model matched the search filter.
        """
        return super().best_model

    @property
    def advanced_models_list(self):
        raise NotImplementedError

    @property
    def normal_models_list(self):
        raise NotImplementedError

    @property
    def simple_models_list(self):
        raise NotImplementedError

    @property
    def dropped_features(self):
        raise NotImplementedError

    @property
    def drop_list(self) -> List[str]:
        raise NotImplementedError

    @property
    def feature_list(self):
        raise NotImplementedError

    def get_cv(self):
        raise NotImplementedError

    def predict_from_dataset(self,
                             dataset,
                             dataset_folder=None) -> pd.DataFrame:
        """ Get the predictions for a dataset stored in the current active [client]
        workspace using the best model of the experiment.

        Arguments:
            dataset (:class:`.Dataset`): Reference to the dataset object to make
                predictions for
            dataset_folder (:class:`.Dataset`): Matching folder dataset for the
                predictions, if necessary

        Returns:
            ``pd.DataFrame``: Predictions as a ``pandas`` dataframe
        """
        return super().predict_from_dataset(dataset, confidence=False, dataset_folder=dataset_folder)

    def predict(self, df, prediction_dataset_name=None) -> pd.DataFrame:
        """ Get the predictions for a dataset stored in the current active [client]
        workspace using the best model of the experiment with a Scikit-learn style blocking prediction mode.

        .. warning::

            For large dataframes and complex (blend) models, this can be slow (up to 1-2 hours).
            Prefer using this for simple models and small dataframes, or use option ``use_best_single = True``.

        Args:
            df (``pd.DataFrame``): ``pandas`` DataFrame containing the test data

        Returns:
            tuple(pd.DataFrame, str): Prediction data (as ``pandas`` dataframe) and prediction job ID.
        """
        return super().predict(df=df, confidence=False, prediction_dataset_name=prediction_dataset_name)
